# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sys

import datetime
import mock
from oslo_utils import uuidutils

sys.modules['ncclient'] = mock.MagicMock()
from networking_cisco.plugins.cisco.cfg_agent import device_status
from neutron.tests import base

import networking_cisco.plugins.cisco.common.cisco_constants as cc
_uuid = uuidutils.generate_uuid

TYPE_STRING = 'string'
TYPE_DATETIME = 'datetime'
NOW = 0
BOOT_TIME = 420
DEAD_TIME = 300
BELOW_BOOT_TIME = 100


def create_timestamp(seconds_from_now, type=TYPE_STRING):
    timedelta = datetime.timedelta(seconds=seconds_from_now)
    past_time = datetime.datetime.utcnow() - timedelta
    if type is TYPE_STRING:
        return past_time.strftime("%Y-%m-%dT%H:%M:%S.%f")
    if type is TYPE_DATETIME:
        return past_time


class TestHostingDevice(base.BaseTestCase):

    def setUp(self):
        super(TestHostingDevice, self).setUp()
        self.status = device_status.DeviceStatus()
        device_status._is_pingable = mock.MagicMock(return_value=True)

        self.hosting_device = {'id': 123,
                               'host_type': 'VM',
                               'management_ip_address': '10.0.0.1',
                               'protcol_port': '22',
                               'booting_time': 420}
        self.created_at_str = datetime.datetime.utcnow().strftime(
            "%Y-%m-%d %H:%M:%S")
        self.hosting_device['created_at'] = self.created_at_str

        # if is_pingable, then hd_state is 'Active'
        self.hosting_device['hd_state'] = cc.HD_ACTIVE
        self.router_id = _uuid()
        self.router = {id: self.router_id,
                       'hosting_device': self.hosting_device}

    def test_hosting_devices_object(self):
        self.assertEqual({}, self.status.hosting_devices_backlog)

    def test_is_hosting_device_reachable_positive(self):
        self.assertTrue(self.status.is_hosting_device_reachable(
            self.hosting_device))

    def test_is_hosting_device_reachable_positive_heartbeat_disabled(self):
        self.status.enable_heartbeat = False
        self.assertTrue(self.status.is_hosting_device_reachable(
            self.hosting_device))
        self.assertEqual(0, len(self.status.get_backlogged_hosting_devices()))

    def test_is_hosting_device_reachable_positive_heartbeat_enabled(self):
        self.status.enable_heartbeat = True
        self.assertTrue(self.status.is_hosting_device_reachable(
            self.hosting_device))
        self.assertEqual(1, len(self.status.get_backlogged_hosting_devices()))
        self.assertTrue(123 in self.status.get_backlogged_hosting_devices())
        self.assertEqual(self.status.hosting_devices_backlog[123]['hd'],
                         self.hosting_device)

    def test_is_hosting_device_reachable_negative(self):
        self.assertEqual(0, len(self.status.hosting_devices_backlog))
        self.hosting_device['created_at'] = self.created_at_str  # Back to str
        device_status._is_pingable.return_value = False
        self.hosting_device['hd_state'] = cc.HD_NOT_RESPONDING

        self.assertFalse(device_status._is_pingable('1.2.3.4'))
        self.assertFalse(self.status.is_hosting_device_reachable(
            self.hosting_device))
        self.assertEqual(1, len(self.status.get_backlogged_hosting_devices()))
        self.assertTrue(123 in self.status.get_backlogged_hosting_devices())
        self.assertEqual(self.status.hosting_devices_backlog[123]['hd'],
                         self.hosting_device)

    def test_is_hosting_device_reachable_negative_heartbeat_disabled(self):
        """
        Even if heartbeat is disabled, unreachable hosting device should
        still be backlogged
        """
        self.status.enable_heartbeat = False

        self.assertEqual(0, len(self.status.hosting_devices_backlog))
        self.hosting_device['created_at'] = self.created_at_str  # Back to str
        device_status._is_pingable.return_value = False
        self.hosting_device['hd_state'] = cc.HD_NOT_RESPONDING

        self.assertFalse(device_status._is_pingable('1.2.3.4'))
        self.assertFalse(self.status.is_hosting_device_reachable(
            self.hosting_device))
        self.assertEqual(1, len(self.status.get_backlogged_hosting_devices()))
        self.assertTrue(123 in self.status.get_backlogged_hosting_devices())
        self.assertEqual(self.status.hosting_devices_backlog[123]['hd'],
                         self.hosting_device)

    def test_test_is_hosting_device_reachable_negative_exisiting_hd(self):
        self.status.hosting_devices_backlog.clear()
        self.status.hosting_devices_backlog[123] = {'hd': self.hosting_device}

        self.assertEqual(1, len(self.status.hosting_devices_backlog))
        self.assertEqual(True, self.status.is_hosting_device_reachable(
            self.hosting_device))
        self.assertEqual(1, len(self.status.get_backlogged_hosting_devices()))
        self.assertTrue(123 in self.status.hosting_devices_backlog.keys())
        self.assertEqual(self.status.hosting_devices_backlog[123]['hd'],
                         self.hosting_device)

    def test_is_hosting_device_reachable_negative_dead_hd(self):
        self.status.hosting_devices_backlog.clear()
        self.hosting_device['hd_state'] = cc.HD_DEAD
        self.status.hosting_devices_backlog = {
            self.hosting_device['id']: {'hd': self.hosting_device}
        }
        self.assertEqual(False, self.status.is_hosting_device_reachable(
            self.hosting_device))
        self.hosting_device['hd_state'] = cc.HD_ACTIVE
        self.status.hosting_devices_backlog.clear()


class TestBacklogHostingDevice(base.BaseTestCase):
    """
    This test class covers test cases pertaining to hosting device backlog
    handling in the DeviceStatus class.  The "backlog" represents a set
    of hosting-devices that should be monitored for liveness and is used
    to retry router update scenarios and for heartbeat / device syncing.
    """

    def setUp(self):
        super(TestBacklogHostingDevice, self).setUp()

        self.status = device_status.DeviceStatus()
        device_status._is_pingable = mock.MagicMock(return_value=True)
        self.can_connect_mock = mock.MagicMock(return_value=True)
        device_status._can_connect = self.can_connect_mock

        self.hosting_device = {'id': 123,
                               'host_type': 'VM',
                               'management_ip_address': '10.0.0.1',
                               'protocol_port': '22',
                               'booting_time': 420}
        self.created_at_str = datetime.datetime.utcnow().strftime(
            "%Y-%m-%d %H:%M:%S")
        self.hosting_device['created_at'] = self.created_at_str

        # if is_pingable, then hd_state is 'Active'
        self.hosting_device['hd_state'] = cc.HD_ACTIVE
        self.router_id = _uuid()
        self.router = {id: self.router_id,
                       'hosting_device': self.hosting_device}

        # mock out drv_mgr
        self.drv_mgr = mock.MagicMock()
        self.drv_mock = mock.MagicMock()
        self.drv_mgr.get_driver_for_hosting_device.return_value = self.drv_mock

    def test_check_backlog_empty(self):

        expected = {'reachable': [],
                    'revived': [],
                    'dead': []}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)

    def test_check_backlog_below_booting_time(self):
        expected = {'reachable': [],
                    'revived': [],
                    'dead': []}

        self.hosting_device['created_at'] = create_timestamp(NOW)
        hd = self.hosting_device
        hd_id = hd['id']
        hd['hd_state'] = cc.HD_NOT_RESPONDING
        self.status.hosting_devices_backlog[hd_id] = {
            'hd': hd,
            'routers': [self.router_id]}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)

        # Simulate 20 seconds before boot time finishes
        self.hosting_device['created_at'] = create_timestamp(BOOT_TIME - 20)
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)

        # Simulate 1 second before boot time
        self.hosting_device['created_at'] = create_timestamp(BOOT_TIME - 1)
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))

    def test_check_backlog_above_booting_time_pingable(self):
        """Test for backlog processing after booting.

        Simulates a hosting device which has passed the created time.
        The device should now be pingable.
        """
        self.hosting_device['created_at'] = create_timestamp(BOOT_TIME + 10)
        hd = self.hosting_device
        hd_id = hd['id']
        device_status._is_pingable.return_value = True
        # assumption in this scenario was that reachability to the
        # hosting-device was unknown
        hd['hd_state'] = cc.HD_NOT_RESPONDING
        self.status.hosting_devices_backlog[hd_id] = {
            'hd': hd,
            'routers': [self.router_id]}
        expected = {'reachable': [hd_id],
                    'revived': [],
                    'dead': []}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.drv_mgr.call_count)

    def test_check_backlog_above_BT_not_pingable_below_deadtime(self):
        """Test for backlog processing in dead time interval.

        This test simulates a hosting device which has passed the created
        time but less than the 'declared dead' time.
        Hosting device is still not pingable.
        """
        hd = self.hosting_device
        hd['created_at'] = create_timestamp(BOOT_TIME + 10)
        #Inserted in backlog now
        hd['backlog_insertion_ts'] = create_timestamp(NOW, type=TYPE_DATETIME)
        hd_id = hd['id']
        device_status._is_pingable.return_value = False
        hd['hd_state'] = cc.HD_NOT_RESPONDING
        self.status.hosting_devices_backlog[hd_id] = {
            'hd': hd,
            'routers': [self.router_id]}
        expected = {'reachable': [],
                    'revived': [],
                    'dead': []}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)

    def test_check_backlog_above_BT_not_pingable_aboveDeadTime(self):
        """Test for backlog processing after dead time interval.

        This test simulates a hosting device which has passed the
        created time but greater than the 'declared dead' time.
        Hosting device is still not pingable.
        """
        hd = self.hosting_device
        hd['created_at'] = create_timestamp(BOOT_TIME + DEAD_TIME + 10)
        #Inserted in backlog 5 seconds after booting time
        hd['backlog_insertion_ts'] = create_timestamp(BOOT_TIME + 5,
                                                      type=TYPE_DATETIME)

        hd_id = hd['id']
        device_status._is_pingable.return_value = False
        hd['hd_state'] = cc.HD_NOT_RESPONDING
        self.status.hosting_devices_backlog[hd_id] = {
            'hd': hd,
            'routers': [self.router_id]}
        expected = {'reachable': [],
                    'revived': [],
                    'dead': [hd_id]}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)
        post_hd_state = (
            self.status.hosting_devices_backlog[hd_id]['hd']['hd_state'])
        self.assertEqual(cc.HD_DEAD, post_hd_state)

    def test_check_backlog_above_BT_revived_hosting_device(self):
        """
        Test reviving a hosting device after it's been deemed dead

        This test simulates a hosting device which has died is now
        reachable again.
        """
        hd = self.hosting_device
        hd['created_at'] = create_timestamp(BOOT_TIME + DEAD_TIME + 10)
        hd['backlog_insertion_ts'] = create_timestamp(BOOT_TIME + 5,
                                                      type=TYPE_DATETIME)
        hd_id = hd['id']
        device_status._is_pingable.return_value = False
        hd['hd_state'] = cc.HD_NOT_RESPONDING
        self.status.hosting_devices_backlog[hd_id] = {
            'hd': hd,
            'routers': [self.router_id]}
        expected = {'reachable': [],
                    'revived': [],
                    'dead': [hd_id]}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)

        post_hd_state = (
            self.status.hosting_devices_backlog[hd_id]['hd']['hd_state'])
        self.assertEqual(cc.HD_DEAD, post_hd_state)

        # now simulate that the hosting device is resurrected
        self.assertEqual(1, len(self.status.get_backlogged_hosting_devices()))
        device_status._is_pingable.return_value = True

        expected = {'reachable': [],
                    'revived': [hd_id],
                    'dead': []}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(1, self.can_connect_mock.call_count)

        post_hd_state = (
            self.status.hosting_devices_backlog[hd_id]['hd']['hd_state'])
        self.assertEqual(cc.HD_ACTIVE, post_hd_state)

    def test_check_backlog_above_BT_reachable_hosting_device(self):
        """
        Test reviving a hosting device after it's been deemed
        unresponsive and then becomes pingable before it's deemed
        dead.
        """
        hd = self.hosting_device
        hd['created_at'] = create_timestamp(BOOT_TIME + DEAD_TIME + 10)
        hd['backlog_insertion_ts'] = create_timestamp(BOOT_TIME + 5,
                                                      type=TYPE_DATETIME)
        hd_id = hd['id']
        device_status._is_pingable.return_value = True
        hd['hd_state'] = cc.HD_NOT_RESPONDING
        self.status.hosting_devices_backlog[hd_id] = {
            'hd': hd,
            'routers': [self.router_id]}

        expected = {'reachable': [hd_id],
                    'revived': [],
                    'dead': []}
        self.assertEqual(
            expected,
            self.status.check_backlogged_hosting_devices(self.drv_mgr))
        self.assertEqual(0, self.can_connect_mock.call_count)

        post_hd_state = (
            self.status.hosting_devices_backlog[hd_id]['hd']['hd_state'])
        self.assertEqual(cc.HD_ACTIVE, post_hd_state)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

from mox3 import mox

from neutronclient.tests.unit import test_cli20

from networking_cisco.neutronclient import hostingdevicescheduler as scheduler

AGENT_ID = 'agent_id1'
HOSTING_DEVICE_ID = 'hd_id1'


class CLITestV20L3HostingDeviceConfigAgentScheduler(test_cli20.CLITestV20Base):

    def _test_assoc_with_cfg_agent(self, resource, cmd, cmd_args, destination,
                                   body, result):
        path = ((scheduler.ConfigAgentHandlingHostingDevice.resource_path +
                 destination) % cmd_args[0])
        self.mox.StubOutWithMock(cmd, "get_client")
        self.mox.StubOutWithMock(self.client.httpclient, "request")
        cmd.get_client().MultipleTimes().AndReturn(self.client)
        result_str = self.client.serialize(result)
        return_tup = (test_cli20.MyResp(200), result_str)

        self.client.httpclient.request(
            test_cli20.end_url(path), 'POST',
            body=test_cli20.MyComparator(body, self.client),
            headers=mox.ContainsKeyValue(
                'X-Auth-Token', test_cli20.TOKEN)).AndReturn(return_tup)
        self.mox.ReplayAll()
        cmd_parser = cmd.get_parser('test_' + resource)
        parsed_args = cmd_parser.parse_args(cmd_args)
        cmd.run(parsed_args)
        self.mox.VerifyAll()
        self.mox.UnsetStubs()

    def _test_disassoc_with_cfg_agent(self, resource, cmd, cmd_args,
                                      destination):
        path = ((scheduler.ConfigAgentHandlingHostingDevice.resource_path +
                 destination + '/%s') % cmd_args)
        self.mox.StubOutWithMock(cmd, "get_client")
        self.mox.StubOutWithMock(self.client.httpclient, "request")
        cmd.get_client().MultipleTimes().AndReturn(self.client)

        return_tup = (test_cli20.MyResp(204), None)
        self.client.httpclient.request(
            test_cli20.end_url(path), 'DELETE',
            body=None,
            headers=mox.ContainsKeyValue(
                'X-Auth-Token', test_cli20.TOKEN)).AndReturn(return_tup)
        self.mox.ReplayAll()
        cmd_parser = cmd.get_parser('test_' + resource)
        parsed_args = cmd_parser.parse_args(cmd_args)
        cmd.run(parsed_args)
        self.mox.VerifyAll()
        self.mox.UnsetStubs()

    def test_associate_hosting_device_with_cfg_agent(self):
        resource = 'hosting_device'
        cmd = scheduler.HostingDeviceAssociateWithConfigAgent(
            test_cli20.MyApp(sys.stdout), None)
        args = (AGENT_ID, HOSTING_DEVICE_ID)
        body = {'hosting_device_id': HOSTING_DEVICE_ID}
        result = {}
        self._test_assoc_with_cfg_agent(
            resource, cmd, args, scheduler.CFG_AGENT_HOSTING_DEVICES, body,
            result)

    def test_disassociate_hosting_device_with_cfg_agent(self):
        resource = 'hosting_device'
        cmd = scheduler.HostingDeviceDisassociateFromConfigAgent(
            test_cli20.MyApp(sys.stdout), None)
        args = (AGENT_ID, HOSTING_DEVICE_ID)
        self._test_disassoc_with_cfg_agent(
            resource, cmd, args, scheduler.CFG_AGENT_HOSTING_DEVICES)

    def test_list_hosting_devices_handled_by_cfg_agent(self):
        resources = 'hosting_devices'
        cmd = scheduler.HostingDeviceHandledByConfigAgentList(
            test_cli20.MyApp(sys.stdout), None)
        hosting_device_id = HOSTING_DEVICE_ID
        path = ((scheduler.ConfigAgentHandlingHostingDevice.resource_path +
                 scheduler.CFG_AGENT_HOSTING_DEVICES) % hosting_device_id)
        contents = [{self.id_field: 'myid1', 'name': 'my_name',
                     'description': 'A fast one', 'admin_state_up': True,
                     'template_id': 'templateid',
                     'management_ip_address': '10.11.12.13'}]
        res = self._test_list_resources(
            resources, cmd, base_args=[hosting_device_id], path=path,
            response_contents=contents)
        self.assertIn('name', res)
        self.assertIn('admin_state_up', res)
        self.assertIn('template_id', res)
        self.assertNotIn('description', res)
        self.assertNotIn('management_ip_address', res)

    def test_list_cfg_agents_handling_hosting_device(self):
        resources = 'agents'
        cmd = scheduler.ConfigAgentHandlingHostingDeviceList(
            test_cli20.MyApp(sys.stdout), None)
        hosting_device_id = HOSTING_DEVICE_ID
        _resource_path = '/dev_mgr/hosting_devices/%s'
        path = ((_resource_path + scheduler.HOSTING_DEVICE_CFG_AGENTS) %
                hosting_device_id)
        contents = [{self.id_field: 'myid1', 'alive': ':-)', 'topic': 'L3Cfg',
                     'agent_type': 'Cisco_Config_agent', 'description': 'None',
                     'admin_state_up': True, 'host': 'Controller',
                     'management_ip_address': '10.11.12.13'}]
        res = self._test_list_resources(
            resources, cmd, base_args=[hosting_device_id], path=path,
            response_contents=contents)
        self.assertIn('alive', res)
        self.assertIn('agent_type', res)
        self.assertIn('admin_state_up', res)
        self.assertIn('host', res)
        self.assertNotIn('description', res)
        self.assertNotIn('topic', res)



# Copyright 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import dfa_sys_lib as utils
from networking_cisco.apps.saf.server import dfa_events_handler as deh

LOG = logging.getLogger(__name__)


class DfaNeutronHelper(object):

    """Helper Routines for Neutron. """

    def __init__(self):
        """Initialization. """
        self.neutron_help = deh.EventsHandler('neutron', None, 20, 25)
        cfg = config.CiscoDFAConfig('neutron').cfg
        self.root_helper = cfg.sys.root_helper

    @property
    def neutronclient(self):
        """Returns client object. """
        return self.neutron_help.nclient

    def create_network(self, name, tenant_id, subnet, gw=None):
        """Create the openstack network, including the subnet. """

        try:
            body = {'network': {'name': name, 'tenant_id': tenant_id,
                                'admin_state_up': True}}
            netw = self.neutronclient.create_network(body=body)
            net_dict = netw.get('network')
            net_id = net_dict.get('id')
        except Exception as exc:
            LOG.error("Failed to create network %(name)s, Exc %(exc)s",
                      {'name': name, 'exc': str(exc)})
            return None, None

        try:
            if gw is None:
                body = {'subnet': {'cidr': subnet,
                                   'ip_version': 4,
                                   'network_id': net_id,
                                   'tenant_id': tenant_id,
                                   'enable_dhcp': False}}
            else:
                body = {'subnet': {'cidr': subnet,
                                   'ip_version': 4,
                                   'network_id': net_id,
                                   'tenant_id': tenant_id,
                                   'enable_dhcp': False,
                                   'gateway_ip': gw}}
            subnet_ret = self.neutronclient.create_subnet(body=body)
            subnet_dict = subnet_ret.get('subnet')
            subnet_id = subnet_dict.get('id')
        except Exception as exc:
            LOG.error("Failed to create subnet %(sub)s, exc %(exc)s",
                      {'sub': subnet, 'exc': str(exc)})
            try:
                self.neutronclient.delete_network(net_id)
            except Exception as exc:
                LOG.error("Failed to delete network %(net)s, exc %(exc)s",
                          {'net': net_id, 'exc': str(exc)})
            return None, None
        return net_id, subnet_id

    def delete_network(self, name, tenant_id, subnet_id, net_id):
        """Delete the openstack subnet and network. """
        try:
            self.neutronclient.delete_subnet(subnet_id)
        except Exception as exc:
            LOG.error("Failed to delete subnet %(sub)s exc %(exc)s",
                      {'sub': subnet_id, 'exc': str(exc)})
            return
        try:
            self.neutronclient.delete_network(net_id)
        except Exception as exc:
            LOG.error("Failed to delete network %(name)s exc %(exc)s",
                      {'name': name, 'exc': str(exc)})

    # Pass
    def delete_network_all_subnets(self, net_id):
        """Delete the openstack network including all its subnets. """
        try:
            body = {'network_id': net_id}
            subnet_list = self.neutronclient.list_subnets(body=body)
            subnet_list = subnet_list.get('subnets')
            for subnet in subnet_list:
                if subnet.get('network_id') == net_id:
                    subnet_id = subnet.get('id')
                    self.neutronclient.delete_subnet(subnet_id)
        except Exception as exc:
            LOG.error("Failed to delete subnet for net %(net)s "
                      "Exc %(exc)s", {'net': net_id, 'exc': str(exc)})
            return False
        try:
            self.neutronclient.delete_network(net_id)
        except Exception as exc:
            LOG.error("Failed to delete network %(net)s Exc %(exc)s",
                      {'net': net_id, 'exc': str(exc)})
            return False
        return True

    def is_subnet_present(self, subnet_addr):
        """Returns if a subnet is present. """
        try:
            subnet_list = self.neutronclient.list_subnets(body={})
            subnet_dat = subnet_list.get('subnets')
            for sub in subnet_dat:
                if sub.get('cidr') == subnet_addr:
                    return True
            return False
        except Exception as exc:
            LOG.error("Failed to list subnet %(sub)s, Exc %(exc)s",
                      {'sub': subnet_addr, 'exc': str(exc)})
            return False

    def get_all_subnets_cidr(self, no_mask=False):
        """Returns all the subnets. """
        body = {}
        subnet_cidrs = []
        try:
            subnet_list = self.neutronclient.list_subnets(body=body)
            subnet_dat = subnet_list.get('subnets')
            for sub in subnet_dat:
                if no_mask:
                    subnet_cidrs.append(sub.get('cidr').split('/')[0])
                else:
                    subnet_cidrs.append(sub.get('cidr'))
        except Exception as exc:
            LOG.error("Failed to list subnet Exc %s", str(exc))
        return subnet_cidrs

    def get_subnets_for_net(self, net):
        """Returns the subnets in a network. """
        try:
            subnet_list = self.neutronclient.list_subnets(network_id=net)
            subnet_dat = subnet_list.get('subnets')
            return subnet_dat
        except Exception as exc:
            LOG.error("Failed to list subnet net %(net)s, Exc: %(exc)s",
                      {'net': net, 'exc': str(exc)})
            return None

    def get_subnet_cidr(self, subnet_id):
        """retrieve the CIDR associated with a subnet, given its ID. """
        try:
            subnet_list = self.neutronclient.list_subnets(id=subnet_id)
            subnet_dat = subnet_list.get('subnets')[0]
            return subnet_dat.get('cidr')
        except Exception as exc:
            LOG.error("Failed to list subnet for ID %(subnet)s, "
                      "exc %(exc)s", {'subnet': subnet_id, 'exc': exc})
            return None

    def delete_network_subname(self, sub_name):
        """Delete the network by part of its name, use with caution. """
        try:
            body = {}
            net_list = self.neutronclient.list_networks(body=body)
            for net in net_list:
                if net.get('name').find(sub_name) != -1:
                    self.delete_network_all_subnets(net.get('net_id'))
        except Exception as exc:
            LOG.error("Failed to get network by subname %(name)s, "
                      "Exc %(exc)s",
                      {'name': sub_name, 'exc': str(exc)})

    def get_network_by_name(self, nwk_name):
        """Search for a openstack network by name. """
        ret_net_lst = []
        try:
            body = {}
            net_list = self.neutronclient.list_networks(body=body)
            net_list = net_list.get('networks')
            for net in net_list:
                if net.get('name') == nwk_name:
                    ret_net_lst.append(net)
        except Exception as exc:
            LOG.error("Failed to get network by name %(name)s, "
                      "Exc %(exc)s",
                      {'name': nwk_name, 'exc': str(exc)})
        return ret_net_lst

    def get_network_by_tenant(self, tenant_id):
        """Returns the network of a given tenant. """
        ret_net_lst = []
        try:
            net_list = self.neutronclient.list_networks(body={})
            for net in net_list.get('networks'):
                if net.get('tenant_id') == tenant_id:
                    ret_net_lst.append(net)
        except Exception as exc:
            LOG.error("Failed to get network by tenant %(tenant)s, "
                      "Exc %(exc)s",
                      {'tenant': tenant_id, 'exc': str(exc)})
        return ret_net_lst

    # Tested
    def get_rtr_by_name(self, rtr_name):
        """Search a router by its name. """
        upd_rtr_list = []
        try:
            rtr_list = self.neutronclient.list_routers()
            for rtr in rtr_list.get('routers'):
                if rtr_name == rtr['name']:
                    upd_rtr_list.append(rtr)
        except Exception as exc:
            LOG.error("Failed to get router by name %(name)s, "
                      "Exc %(exc)s",
                      {'name': rtr_name, 'exc': str(exc)})
        return upd_rtr_list

    def create_router(self, name, tenant_id, subnet_lst):
        """Create a openstack router and add the interfaces. """
        try:
            body = {'router': {'name': name, 'tenant_id': tenant_id,
                               'admin_state_up': True}}
            router = self.neutronclient.create_router(body=body)
            rout_dict = router.get('router')
            rout_id = rout_dict.get('id')
        except Exception as exc:
            LOG.error("Failed to create router with name %(name)s"
                      " Exc %(exc)s", {'name': name, 'exc': str(exc)})
            return None

        ret = self.add_intf_router(rout_id, tenant_id, subnet_lst)
        if not ret:
            try:
                ret = self.neutronclient.delete_router(rout_id)
            except Exception as exc:
                LOG.error("Failed to delete router %(name)s, Exc %(exc)s",
                          {'name': name, 'exc': str(exc)})
            return None
        return rout_id

    def add_intf_router(self, rout_id, tenant_id, subnet_lst):
        """Add the interfaces to a router. """
        try:
            for subnet_id in subnet_lst:
                body = {'subnet_id': subnet_id}
                intf = self.neutronclient.add_interface_router(rout_id,
                                                               body=body)
                intf.get('port_id')
        except Exception as exc:
            LOG.error("Failed to create router intf ID %(id)s,"
                      " Exc %(exc)s", {'id': rout_id, 'exc': str(exc)})
            return False
        return True

    # Passed
    def delete_router(self, name, tenant_id, rout_id, subnet_lst):
        """Delete the openstack router.

        Delete the router and remove the interfaces attached to it.
        """
        ret = self.delete_intf_router(name, tenant_id, rout_id, subnet_lst)
        if not ret:
            return False

        try:
            ret = self.neutronclient.delete_router(rout_id)
        except Exception as exc:
            LOG.error("Failed to delete router %(name)s ret %(ret)s "
                      "Exc %(exc)s",
                      {'name': name, 'ret': str(ret), 'exc': str(exc)})
            return False
        return True

    def delete_intf_router(self, name, tenant_id, rout_id, subnet_lst):
        """Delete the openstack router and remove the interfaces attached. """
        try:
            for subnet_id in subnet_lst:
                body = {'subnet_id': subnet_id}
                intf = self.neutronclient.remove_interface_router(rout_id,
                                                                  body=body)
                intf.get('id')
        except Exception as exc:
            LOG.error("Failed to delete router interface %(name)s, "
                      " Exc %(exc)s", {'name': name, 'exc': str(exc)})
            return False
        return True

    def delete_router_by_name(self, rtr_name, tenant_id):
        """Delete the openstack router and its interfaces given its name.

        The interfaces should be already removed prior to calling this
        function.
        """
        try:
            routers = self.neutronclient.list_routers()
            rtr_list = routers.get('routers')
            for rtr in rtr_list:
                if rtr_name == rtr['name']:
                    self.neutronclient.delete_router(rtr['id'])
        except Exception as exc:
            LOG.error("Failed to get and delete router by name %(name)s, "
                      "Exc %(exc)s",
                      {'name': rtr_name, 'exc': str(exc)})
            return False
        return True

    def get_router_intf(self, router_id):
        """Retrieve the router interfaces. Incomplete, TODO(padkrish). """
        try:
            body = {}
            self.neutronclient.show_router(router_id, body=body)
        except Exception as exc:
            LOG.error("Failed to show router interface %(id)s "
                      "Exc %(exc)s", {'id': router_id, 'exc': str(exc)})
            return
        # Complete fixme(padkrish)

    def get_router_port_subnet(self, subnet_id):
        try:
            body = 'network:router_interface'
            port_data = self.neutronclient.list_ports(device_owner=body)
            port_list = port_data.get('ports')
            for port in port_list:
                sub = port.get('fixed_ips')[0].get('subnet_id')
                if sub == subnet_id:
                    return port
            return None
        except Exception as exc:
            LOG.error("Failed to get router port subnet %(net)s, "
                      "Exc: %(exc)s", {'net': subnet_id, 'exc': str(exc)})
            return None

    def get_rtr_name(self, router_id):
        """Retrieve the router name. Incomplete. """
        try:
            body = {}
            router = self.neutronclient.show_router(router_id, body=body)
            return router.get('router').get('name')
        except Exception as exc:
            LOG.error("Failed to show router interface %(id)s "
                      "Exc %(exc)s", {'id': router_id, 'exc': str(exc)})

    def find_rtr_namespace(self, rout_id):
        """Find the namespace associated with the router. """
        if rout_id is None:
            return None
        args = ['ip', 'netns', 'list']
        try:
            ns_list = utils.execute(args, root_helper=self.root_helper)
        except Exception as exc:
            LOG.error("Unable to find the namespace list Exception %s",
                      exc)
            return None
        for ns in ns_list.split():
            if 'router' in ns and rout_id in ns:
                return ns

    def program_rtr(self, args, rout_id, namespace=None):
        """Execute the command against the namespace. """
        if namespace is None:
            namespace = self.find_rtr_namespace(rout_id)
        if namespace is None:
            LOG.error("Unable to find namespace for router %s", rout_id)
            return False
        final_args = ['ip', 'netns', 'exec', namespace] + args
        try:
            utils.execute(final_args, root_helper=self.root_helper)
        except Exception as e:
            LOG.error("Unable to execute %(cmd)s. "
                      "Exception: %(exception)s",
                      {'cmd': final_args, 'exception': e})
            return False
        return True

    def program_rtr_return(self, args, rout_id, namespace=None):
        """Execute the command against the namespace and return the result. """
        if namespace is None:
            namespace = self.find_rtr_namespace(rout_id)
        if namespace is None:
            LOG.error("Unable to find namespace for router %s", rout_id)
            return False
        final_args = ['ip', 'netns', 'exec', namespace] + args
        try:
            return utils.execute(final_args, root_helper=self.root_helper)
        except Exception as e:
            LOG.error("Unable to execute %(cmd)s. "
                      "Exception: %(exception)s",
                      {'cmd': final_args, 'exception': e})
            return None

    def program_rtr_default_gw(self, tenant_id, rout_id, gw):
        """Program the default gateway of a router. """
        args = ['route', 'add', 'default', 'gw', gw]
        ret = self.program_rtr(args, rout_id)
        if not ret:
            LOG.error("Program router returned error for %s", rout_id)
            return False
        return True

    def get_subnet_nwk_excl(self, tenant_id, excl_list, excl_part=False):
        """Retrieve the subnets of a network.

        Get the subnets inside a network after applying the exclusion
        list.
        """
        net_list = self.get_network_by_tenant(tenant_id)
        ret_subnet_list = []
        for net in net_list:
            if excl_part:
                name = net.get('name')
                part = name.partition('::')[2]
                if part:
                    continue
            subnet_lst = self.get_subnets_for_net(net.get('id'))
            for subnet_elem in subnet_lst:
                subnet = subnet_elem.get('cidr').split('/')[0]
                subnet_and_mask = subnet_elem.get('cidr')
                if subnet not in excl_list:
                    ret_subnet_list.append(subnet_and_mask)
        return ret_subnet_list

    def program_rtr_all_nwk_next_hop(self, tenant_id, rout_id, next_hop,
                                     excl_list):
        """Program the next hop for all networks of a tenant. """
        namespace = self.find_rtr_namespace(rout_id)
        if namespace is None:
            LOG.error("Unable to find namespace for router %s", rout_id)
            return False

        net_list = self.get_network_by_tenant(tenant_id)
        for net in net_list:
            subnet_lst = self.get_subnets_for_net(net.get('id'))
            for subnet_elem in subnet_lst:
                subnet = subnet_elem.get('cidr').split('/')[0]
                subnet_and_mask = subnet_elem.get('cidr')
                if subnet not in excl_list:
                    args = ['route', 'add', '-net', subnet_and_mask, 'gw',
                            next_hop]
                    ret = self.program_rtr(args, rout_id, namespace=namespace)
                    if not ret:
                        LOG.error("Program router returned error for %s",
                                  rout_id)
                        return False
        return True

    def program_rtr_nwk_next_hop(self, rout_id, next_hop, cidr):
        """Program the next hop for all networks of a tenant. """
        namespace = self.find_rtr_namespace(rout_id)
        if namespace is None:
            LOG.error("Unable to find namespace for router %s", rout_id)
            return False

        args = ['route', 'add', '-net', cidr, 'gw', next_hop]
        ret = self.program_rtr(args, rout_id, namespace=namespace)
        if not ret:
            LOG.error("Program router returned error for %s", rout_id)
            return False
        return True

    def remove_rtr_nwk_next_hop(self, rout_id, next_hop, subnet_lst,
                                excl_list):
        """Remove the next hop for all networks of a tenant. """
        namespace = self.find_rtr_namespace(rout_id)
        if namespace is None:
            LOG.error("Unable to find namespace for router %s", rout_id)
            return False

        args = ['ip', 'route']
        ret = self.program_rtr_return(args, rout_id, namespace=namespace)
        if ret is None:
            LOG.error("Get routes return None %s", rout_id)
            return False
        routes = ret.split('\n')
        concat_lst = subnet_lst + excl_list
        for rout in routes:
            if len(rout) == 0:
                continue
            nwk = rout.split()[0]
            if nwk == 'default':
                continue
            nwk_no_mask = nwk.split('/')[0]
            if nwk_no_mask not in concat_lst and nwk not in concat_lst:
                args = ['route', 'del', '-net', nwk, 'gw', next_hop]
                ret = self.program_rtr(args, rout_id, namespace=namespace)
                if not ret:
                    LOG.error("Program router returned error for %s",
                              rout_id)
                    return False
        return True

    def get_fw(self, fw_id):
        """Return the Firewall given its ID. """
        fw = None
        try:
            fw = self.neutronclient.show_firewall(fw_id)
        except Exception as exc:
            LOG.error("Failed to get firewall list for id %(id)s, "
                      "Exc %(exc)s", {'id': fw_id, 'exc': str(exc)})
        return fw

    # Tested
    def get_fw_rule(self, rule_id):
        """Return the firewall rule, given its ID. """
        rule = None
        try:
            rule = self.neutronclient.show_firewall_rule(rule_id)
        except Exception as exc:
            LOG.error("Failed to get firewall rule for id %(id)s "
                      "Exc %(exc)s", {'id': rule_id, 'exc': str(exc)})
        return rule

    # Tested
    def get_fw_policy(self, policy_id):
        """Return the firewall policy, given its ID. """
        policy = None
        try:
            policy = self.neutronclient.show_firewall_policy(policy_id)
        except Exception as exc:
            LOG.error("Failed to get firewall plcy for id %(id)s "
                      "Exc %(exc)s",
                      {'id': policy_id, 'exc': str(exc)})
        return policy



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six

from networking_cisco.plugins.cisco.common import cisco_constants as const
from networking_cisco.plugins.cisco.device_manager.rpc import (
    devices_cfgagent_rpc_cb)
from neutron.tests import base


class TestCfgAgentDeviceManagerCallbacks(base.BaseTestCase):

    def setUp(self):
        super(TestCfgAgentDeviceManagerCallbacks, self).setUp()

    def test_report_non_responding_hosting_devices(self):
        dm_plugin = mock.MagicMock()
        cb = devices_cfgagent_rpc_cb.DeviceMgrCfgRpcCallback(dm_plugin)
        ctx = mock.MagicMock()
        host = 'some_host'
        hd_ids = ['hd_id1', 'hd_id2']
        update_status_mock = mock.MagicMock()
        cb.update_hosting_device_status = update_status_mock
        cb.report_non_responding_hosting_devices(ctx, host, hd_ids)
        update_status_mock.assert_called_once_with(
            ctx, host, {const.HD_DEAD: hd_ids})

    def test_report_for_duty_triggers_scheduling(self):
        dm_plugin = mock.MagicMock()
        cb = devices_cfgagent_rpc_cb.DeviceMgrCfgRpcCallback(dm_plugin)
        ctx = mock.MagicMock()
        host = 'some_host'
        cb.register_for_duty(ctx, host)
        dm_plugin.auto_schedule_hosting_devices.assert_called_once_with(ctx,
                                                                        host)

    def _test_update_hosting_device_status(self, status_info):
        dm_plugin = mock.MagicMock()
        cb = devices_cfgagent_rpc_cb.DeviceMgrCfgRpcCallback(dm_plugin)
        ctx = mock.MagicMock()
        host = 'some_host'
        cb.update_hosting_device_status(ctx, host, status_info)
        non_resp_calls = []
        for status, hd_ids in six.iteritems(status_info):
            hd_spec = {'hosting_device': {'status': status}}
            update_calls = [mock.call(ctx, hd_id, hd_spec)
                            for hd_id in hd_ids]
            dm_plugin.update_hosting_device.assert_has_calls(update_calls,
                                                             any_order=True)

            if status == const.HD_DEAD or status == const.HD_ERROR:
                non_resp_calls.append(mock.call(ctx, host, hd_ids))
        dm_plugin.handle_non_responding_hosting_devices.has_calls(
            non_resp_calls, any_order=True)
        self.assertEqual(
            len(non_resp_calls),
            dm_plugin.handle_non_responding_hosting_devices.call_count)

    def test_update_hosting_device_status_to_dead(self):
        self._test_update_hosting_device_status(
            {const.HD_DEAD: ['hd_id1', 'hd_id2']})

    def test_update_hosting_device_status_to_active(self):
        self._test_update_hosting_device_status(
            {const.HD_ACTIVE: ['hd_id1', 'hd_id2']})

    def test_update_hosting_device_status_to_error(self):
        self._test_update_hosting_device_status(
            {const.HD_ERROR: ['hd_id1', 'hd_id2']})

    def test_update_hosting_device_status_to_not_responding(self):
        self._test_update_hosting_device_status(
            {const.HD_NOT_RESPONDING: ['hd_id1', 'hd_id2']})

    def test_update_hosting_device_status_multiple(self):
        self._test_update_hosting_device_status(
            {const.HD_ACTIVE: ['hd_id1', 'hd_id2'],
             const.HD_NOT_RESPONDING: ['hd_id3', 'hd_id4']})

    def test_update_hosting_device_status_all(self):
        self._test_update_hosting_device_status(
            {const.HD_ACTIVE: ['hd_id1', 'hd_id2'],
             const.HD_NOT_RESPONDING: ['hd_id3', 'hd_id4', 'hd_id5'],
             const.HD_DEAD: ['hd_id6'],
             const.HD_ERROR: ['hd_id7', 'hd_id8']})

    def test_get_hosting_devices_for_agent(self):
        dm_plugin = mock.MagicMock()
        agent = mock.MagicMock()
        agent.id = 'fake_id'
        dm_plugin.get_cfg_agents.return_value = [agent]
        hds = [{'id': 'fake_id1'}, {'id': 'fake_id2'}]
        ips = {hds[0]['id']: '10.0.0.5/24', hds[1]['id']: '10.0.10.6/24'}
        return_value = [{'id': hd['id'],
                         'management_ip_address': ips[hd['id']]}
                        for hd in hds]
        dm_plugin.get_device_info_for_agent.side_effect = (
            lambda ctxt, hd_db: {'id': hd_db['id'],
                                 'management_ip_address': ips[hd_db['id']]})
        dm_plugin.get_hosting_devices_db.return_value = hds
        cb = devices_cfgagent_rpc_cb.DeviceMgrCfgRpcCallback(dm_plugin)
        ctx = mock.MagicMock()
        host = 'some_host'
        self.assertEqual(return_value,
                         cb.get_hosting_devices_for_agent(ctx, host))



# Copyright (c) 2017-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Implements REST API Client For Nexus
"""

import netaddr
import requests

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    exceptions as cexc)
from oslo_log import log as logging
from oslo_serialization import jsonutils

DEFAULT_HEADER = {"Content-type": "application/json", "Accept": "text/plain"}
COOKIE_HEADER = """
{"Cookie": %s, "Content-type": "application/json", "Accept": "text/plain"}"""
DEFAULT_SCHEME = "https"
ACCEPTED_CODES = [200, 201, 204]
CREDENTIAL_EXPIRED = [403]

LOG = logging.getLogger(__name__)


class CiscoNexusRestapiClient(object):

    def __init__(self, credentials,
                 accepted_codes=ACCEPTED_CODES,
                 scheme=DEFAULT_SCHEME,
                 timeout=30,
                 max_retries=2,
                 request_cookie=True):
        """Initialize the rest api client for Nexus."""
        self.format = 'json'
        self.accepted_codes = accepted_codes
        self.action_prefix = DEFAULT_SCHEME + '://%s/'
        self.scheme = scheme
        self.status = requests.codes.OK
        self.time_stats = {}
        self.timeout = timeout
        self.max_retries = max_retries
        self.session = requests.Session()
        self.credentials = credentials
        self.request_cookie = request_cookie

    def _get_cookie(self, mgmt_ip, config, refresh=False):
        """Performs authentication and retries cookie."""

        if mgmt_ip not in self.credentials:
            return None

        security_data = self.credentials[mgmt_ip]
        verify = security_data[const.HTTPS_CERT_TUPLE]
        if not verify:
            verify = security_data[const.HTTPS_VERIFY_TUPLE]

        if not refresh and security_data[const.COOKIE_TUPLE]:
            return security_data[const.COOKIE_TUPLE], verify

        payload = {"aaaUser": {"attributes": {
                   "name": security_data[const.UNAME_TUPLE],
                   "pwd": security_data[const.PW_TUPLE]}}}
        headers = {"Content-type": "application/json", "Accept": "text/plain"}

        url = "{0}://{1}/api/aaaLogin.json".format(DEFAULT_SCHEME, mgmt_ip)

        try:
            response = self.session.request('POST',
                           url,
                           data=jsonutils.dumps(payload),
                           headers=headers,
                           verify=verify,
                           timeout=self.timeout * 2)
        except Exception as e:
            raise cexc.NexusConnectFailed(nexus_host=mgmt_ip,
                                         exc=e)

        self.status = response.status_code
        if response.status_code == requests.codes.OK:
            cookie = response.headers.get('Set-Cookie')
            security_data = (
                security_data[const.UNAME_TUPLE:const.COOKIE_TUPLE] +
                (cookie,))
            self.credentials[mgmt_ip] = security_data
            return cookie, verify
        else:
            e = "REST API connect returned Error code: "
            e += str(self.status)
            raise cexc.NexusConnectFailed(nexus_host=mgmt_ip,
                                         exc=e)

    def send_request(self, method, action, body=None,
                    headers=None, ipaddr=None):
        """Perform the HTTP request.

        The response is in either JSON format or plain text. A GET method will
        invoke a JSON response while a PUT/POST/DELETE returns message from the
        the server in plain text format.
        Exception is raised when server replies with an INTERNAL SERVER ERROR
        status code (500) i.e. an error has occurred on the server or SERVICE
        UNAVAILABLE (404) i.e. server is not reachable.

        :param method: type of the HTTP request. POST, GET, PUT or DELETE
        :param action: path to which the client makes request
        :param body: dict of arguments which are sent as part of the request
        :param headers: header for the HTTP request
        :param server_ip: server_ip for the HTTP request.
        :returns: JSON or plain text in HTTP response
        """

        action = ''.join([self.scheme, '://%s/', action])
        if netaddr.valid_ipv6(ipaddr):
            # Enclose IPv6 address in [] in the URL
            action = action % ("[%s]" % ipaddr)
        else:
            # IPv4 address
            action = action % ipaddr

        config = action + " : " + body if body else action

        # if cookie needed and one not previously created
        if self.request_cookie:
            cookie, verify = self._get_cookie(ipaddr, config)
            headers = {"Content-type": "application/json",
                       "Accept": "text/plain", "Cookie": cookie}
        else:
            if ipaddr not in self.credentials:
                raise cexc.NexusCredentialNotFound(switch_ip=ipaddr)
            else:
                headers = {'Content-Type': 'application/json'}
                security_data = self.credentials[ipaddr]
                verify = security_data[const.HTTPS_CERT_TUPLE]
                if not verify:
                    verify = security_data[const.HTTPS_VERIFY_TUPLE]
                self.session.auth = (security_data[0], security_data[1])
        if self.status != requests.codes.OK:
            return {}

        for attempt in range(self.max_retries + 1):
            try:
                LOG.debug("[Nexus %(ipaddr)s attempt %(id)s]: Connecting.." %
                         {"ipaddr": ipaddr, "id": attempt})
                response = self.session.request(
                    method,
                    action,
                    data=body,
                    headers=headers,
                    verify=verify,
                    timeout=self.timeout)
                if (self.request_cookie and
                    response.status_code in CREDENTIAL_EXPIRED):
                    # if need new cookie
                    cookie, verify = self._get_cookie(
                        ipaddr, config, refresh=True)
                    headers = {"Content-type": "application/json",
                               "Accept": "text/plain", "Cookie": cookie}
                    continue
            except Exception as e:
                LOG.error(
                    "Exception raised %(err)s for Rest/NXAPI %(cfg)s",
                    {'err': str(e), 'cfg': config})
                raise cexc.NexusConfigFailed(nexus_host=ipaddr,
                                             config=config,
                                             exc=e)
            else:
                break

        status_string = requests.status_codes._codes[response.status_code][0]
        if response.status_code in self.accepted_codes:
            LOG.debug(
                "Good status %(status)s(%(code)d) returned for %(url)s",
                {'status': status_string,
                'code': response.status_code,
                'url': action})
            # 'text/json' used with nxapi else application/json with restapi
            output = {}
            if ('application/json' in response.headers['content-type'] or
                'text/json' in response.headers['content-type']):
                try:
                    output = response.json()
                except Exception:
                    pass
            if 'ins_api' in output:
                # do special nxapi response handling
                try:
                    cli_resp = output['ins_api']['outputs']['output']
                except Exception:
                    cli_resp = []
                # Check results for each command
                for cli in cli_resp:
                    try:
                        status = int((cli['code']))
                    except ValueError:
                        status = 'bad_status %s' % cli['code']
                    if status not in self.accepted_codes:
                        excpt = "ins_api CLI failure occurred "
                        "with cli return code %s" % str(status)
                        raise cexc.NexusConfigFailed(
                            nexus_host=ipaddr, config=config,
                            exc=excpt)
            return output
        else:
            LOG.error(
                "Bad status %(status)s(%(code)d) returned for %(url)s",
                {'status': status_string,
                'code': response.status_code,
                'url': action})
            LOG.error("Response text: %(txt)s",
                      {'txt': response.text})
            raise cexc.NexusConfigFailed(nexus_host=ipaddr,
                                         config=config,
                                         exc=response.text)

    def rest_delete(self, action, ipaddr=None, body=None, headers=None):
        return self.send_request("DELETE", action, body=body,
                               headers=headers, ipaddr=ipaddr)

    def rest_get(self, action, ipaddr, body=None, headers=None):
        return self.send_request("GET", action, body=body,
                               headers=headers, ipaddr=ipaddr)

    def rest_post(self, action, ipaddr=None, body=None, headers=None):
        return self.send_request("POST", action, body=body,
                               headers=headers, ipaddr=ipaddr)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six

from oslo_db import exception as db_exc

from neutron.tests import base

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import l3_exceptions as l3
from networking_cisco.plugins.cisco.l3.rpc import l3_router_cfg_agent_rpc_cb


# NOTE(bobmel): The scheduling functions called by some of the callback
# functions are tested in test_l3_routertype_aware_schedulers.py and in
# test_l3_router_appliance_plugin.py (specifically the class
# L3CfgAgentRouterApplianceTestCase).
class TestCfgAgentL3RouterCallbacks(base.BaseTestCase):

    def setUp(self):
        self.l3_plugin = mock.MagicMock()
        self.cb = l3_router_cfg_agent_rpc_cb.L3RouterCfgRpcCallback(
            self.l3_plugin)
        self.contextMock = mock.MagicMock()
        super(TestCfgAgentL3RouterCallbacks, self).setUp()

    def test_cfg_sync_routers_retries_on_db_errors(self):
        router_ids = ['some_router_id_1', 'some_router_id_2']
        routers = [{'id': router_ids[0]}, {'id': router_ids[1]}]
        host = 'some_host'
        hosting_device_ids = [host]
        self.l3_plugin.list_active_sync_routers_on_hosting_devices = (
            mock.MagicMock(side_effect=[db_exc.DBDeadlock, db_exc.DBDeadlock,
                                        routers]))
        result = self.cb.cfg_sync_routers(self.contextMock, host,
                                          router_ids, hosting_device_ids)
        calls = [mock.call(mock.ANY, host, router_ids, hosting_device_ids)
                 for call in [1, 2, 3]]
        (self.l3_plugin.list_active_sync_routers_on_hosting_devices
         .assert_has_calls(calls))
        self.assertEqual(
            3,
            self.l3_plugin.list_active_sync_routers_on_hosting_devices.
            call_count)
        self.assertEqual(routers, result)

    def test_cfg_sync_routers_missing_scheduling_fcn(self):
        self.l3_plugin.list_active_sync_routers_on_hosting_devices = (
            mock.MagicMock(side_effect=AttributeError))
        result = self.cb.cfg_sync_routers(self.contextMock, 'some_host')
        self.assertEqual([], result)

    def test_cfg_sync_all_hosted_routers_retries_on_db_errors(self):
        routers = [{'id': 'some_router_id_1'}, {'id': 'some_router_id_2'}]
        self.l3_plugin.list_all_routers_on_hosting_devices = mock.MagicMock(
            side_effect=[db_exc.DBDeadlock, db_exc.DBDeadlock, routers])
        result = self.cb.cfg_sync_all_hosted_routers(self.contextMock,
                                                     'some_host')
        self.assertEqual(
            3, self.l3_plugin.list_all_routers_on_hosting_devices.call_count)
        self.assertEqual(routers, result)

    def test_cfg_sync_all_hosted_routers_missing_scheduling_fcn(self):
        self.l3_plugin.list_all_routers_on_hosting_devices = mock.MagicMock(
            side_effect=AttributeError)
        result = self.cb.cfg_sync_all_hosted_routers(self.contextMock,
                                                     'some_host')
        self.assertEqual([], result)

    def test_update_floatingip_statuses_cfg_retries_on_db_errors(self):
        fip_statuses = {'fip_id_1': 'ACTIVE'}
        self.l3_plugin.update_floatingip_status = mock.MagicMock(
            side_effect=[db_exc.DBDeadlock, None])
        self.cb.update_floatingip_statuses_cfg(
            self.contextMock, 'some_router_id', fip_statuses)
        calls = [mock.call(self.contextMock, fip_id, status)
                 for fip_id, status in six.iteritems(fip_statuses)]
        self.l3_plugin.update_floatingip_status.assert_has_calls(calls)
        self.assertEqual(2, self.l3_plugin.update_floatingip_status.call_count)

    def test_update_floatingip_statuses_cfg_ignores_missing_fip(self):
        bad_fip_id = 'id_non_existent_fip'
        fip_statuses = {bad_fip_id: 'ACTIVE'}
        self.l3_plugin.update_floatingip_status = mock.MagicMock(
            side_effect=[l3.FloatingIPNotFound(floatingip_id=bad_fip_id)])
        self.l3_plugin.get_floatingips = mock.MagicMock(return_value=[])
        with mock.patch('networking_cisco.plugins.cisco.l3.rpc'
                        '.l3_router_cfg_agent_rpc_cb.LOG.debug') as log_mock:
            self.cb.update_floatingip_statuses_cfg(
                self.contextMock, 'some_router_id', fip_statuses)
        calls = [mock.call("Floating IP: %s no longer present.", bad_fip_id)]
        log_mock.assert_has_calls(calls, any_order=True)
        self.assertEqual(1, self.l3_plugin.update_floatingip_status.call_count)

    def test_update_floatingip_statuses_cfg_sets_status_down_if_no_router(
            self):
        fip_id = 'fip_id_1'
        status = 'ACTIVE'
        fip_statuses = {fip_id: status}
        self.l3_plugin.update_floatingip_status = mock.MagicMock()
        self.l3_plugin.get_floatingips = mock.MagicMock(return_value=[{
            'id': fip_id, 'router_id': None}])
        self.cb.update_floatingip_statuses_cfg(
            self.contextMock, 'some_router_id', fip_statuses)
        calls = [mock.call(self.contextMock, fip_id, status),
                 mock.call(self.contextMock, fip_id,
                           bc.constants.FLOATINGIP_STATUS_DOWN)]
        self.l3_plugin.update_floatingip_status.assert_has_calls(calls)
        self.assertEqual(2, self.l3_plugin.update_floatingip_status.call_count)

    def test_update_port_statuses_cfg_retries_on_db_errors(self):
        port_ids = ['some_port_id_1', 'some_port_id_2']
        status = 'ACTIVE'
        self.l3_plugin.update_router_port_statuses = mock.MagicMock(
            side_effect=[db_exc.DBDeadlock, db_exc.DBDeadlock, port_ids])
        self.cb.update_port_statuses_cfg(self.contextMock, port_ids, status)
        calls = [mock.call(self.contextMock, port_ids, status)
                 for call in [1, 2, 3]]
        self.assertEqual(3,
                         self.l3_plugin.update_router_port_statuses.call_count)
        self.l3_plugin.update_router_port_statuses.assert_has_calls(calls)



# Copyright 2015 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""This file contains the implementation of OpenStack component of VDP.
VDP is a part of LLDP Agent Daemon (lldpad). For more information on VDP,
pls visit http://www.ieee802.org/1/pages/802.1bg.html
"""

import six

from networking_cisco.apps.saf.agent.vdp import (
    lldpad_constants as vdp_const)
from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import dfa_sys_lib as utils
from networking_cisco.apps.saf.common import utils as sys_utils

LOG = logging.getLogger(__name__)

# When timeout support becomes available in lldpad, this config will be
# enabled fixme(padkrish)
# OPTS = [
#    cfg.IntOpt('lldp_timeout',
#               default='0',
#               help=_('Timeout in seconds for lldptool commands')),
# ]


def enable_lldp(self, port_name, is_ncb=True, is_nb=False):
    """Function to enable LLDP on the interface. """

    if is_ncb:
        self.run_lldptool(["-L", "-i", port_name, "-g", "ncb",
                           "adminStatus=rxtx"])
    if is_nb:
        self.run_lldptool(["-L", "-i", port_name, "-g", "nb",
                           "adminStatus=rxtx"])


class LldpadDriver(object):

    """LLDPad driver class. """

    def __init__(self, port_name, phy_uplink, root_helper, is_ncb=True,
                 is_nb=False):
        """Initialize Routine.

        param port_name: Port where LLDP/EVB needs to be cfgd
        param phy_uplink: Physical Interface
        param root_helper: utility to use when running shell cmds.
        param is_ncb: Should LLDP be cfgd on Nearest Customer Bridge
        param is_nb: Should LLDP be cfgd on Nearest Bridge
        """
        # Re-enable this once support becomes available in lldpad.
        # fixme(padkrish)
        # cfg.CONF.register_opts(OPTS)
        self.port_name = port_name
        self.phy_uplink = phy_uplink
        self.is_ncb = is_ncb
        self.is_nb = is_nb
        self.root_helper = root_helper
        self.mutex_lock = sys_utils.lock()
        self.read_vdp_cfg()
        self.vdp_vif_map = {}
        self.oui_vif_map = {}
        self.enable_lldp()
        sync_timeout_val = int(self.vdp_opts['vdp_sync_timeout'])
        vdp_periodic_task = sys_utils.PeriodicTask(sync_timeout_val,
                                                   self._vdp_refrsh_hndlr)
        self.vdp_periodic_task = vdp_periodic_task
        vdp_periodic_task.run()

    def clear_uplink(self):
        self.phy_uplink = None
        self.port_name = None
        self.vdp_periodic_task.stop()
        del self.vdp_vif_map
        del self.oui_vif_map

    def read_vdp_cfg(self):
        self._cfg = config.CiscoDFAConfig().cfg
        self.vdp_opts = dict()
        self.vdp_opts['mgrid'] = self._cfg.vdp.mgrid2
        self.vdp_opts['typeid'] = self._cfg.vdp.typeid
        self.vdp_opts['typeidver'] = self._cfg.vdp.typeidver
        self.vdp_opts['vsiidfrmt'] = self._cfg.vdp.vsiidfrmt
        self.vdp_opts['hints'] = self._cfg.vdp.hints
        self.vdp_opts['filter'] = self._cfg.vdp.filter
        self.vdp_opts['vdp_sync_timeout'] = self._cfg.vdp.vdp_sync_timeout

    def enable_lldp(self):
        """Function to enable LLDP on the interface. """
        if self.is_ncb:
            self.run_lldptool(["-L", "-i", self.port_name, "-g", "ncb",
                               "adminStatus=rxtx"])
        if self.is_nb:
            self.run_lldptool(["-L", "-i", self.port_name, "-g", "nb",
                               "adminStatus=rxtx"])

    def enable_evb(self):
        """Function to enable EVB on the interface. """
        if self.is_ncb:
            self.run_lldptool(["-T", "-i", self.port_name, "-g", "ncb",
                               "-V", "evb", "enableTx=yes"])
            ret = self.enable_gpid()
            return ret
        else:
            LOG.error("EVB cannot be set on NB")
            return False

    def enable_gpid(self):
        """Function to enable Group ID on the interface.

        This is needed to use the MAC, GID, VID Filter.
        """
        if self.is_ncb:
            self.run_lldptool(["-T", "-i", self.port_name, "-g", "ncb",
                               "-V", "evb", "-c", "evbgpid=yes"])
            return True
        else:
            LOG.error("GPID cannot be set on NB")
            return False

    # fixme(padkrish)
    def _vdp_refrsh_hndlr(self):
        """Periodic refresh of vNIC events to VDP.

        VDP daemon itself has keepalives. This is needed on top of it
        to keep Orchestrator like OpenStack, VDP daemon and the physical
        switch in sync.
        """
        LOG.debug("Refresh handler")
        try:
            if not self.vdp_vif_map:
                LOG.debug("vdp_vif_map not created, returning")
                return
            vdp_vif_map = dict.copy(self.vdp_vif_map)
            oui_vif_map = dict.copy(self.oui_vif_map)
            for key in six.iterkeys(vdp_vif_map):
                lvdp_dict = vdp_vif_map.get(key)
                loui_dict = oui_vif_map.get(key)
                if not lvdp_dict:
                    return
                if not loui_dict:
                    oui_id = ""
                    oui_data = ""
                else:
                    oui_id = loui_dict.get('oui_id')
                    oui_data = loui_dict.get('oui_data')
                with self.mutex_lock:
                    if key in self.vdp_vif_map:
                        LOG.debug("Sending Refresh for VSI %s", lvdp_dict)
                        vdp_vlan, fail_reason = self.send_vdp_assoc(
                            vsiid=lvdp_dict.get('vsiid'),
                            mgrid=lvdp_dict.get('mgrid'),
                            typeid=lvdp_dict.get('typeid'),
                            typeid_ver=lvdp_dict.get('typeid_ver'),
                            vsiid_frmt=lvdp_dict.get('vsiid_frmt'),
                            filter_frmt=lvdp_dict.get('filter_frmt'),
                            gid=lvdp_dict.get('gid'),
                            mac=lvdp_dict.get('mac'),
                            vlan=0, oui_id=oui_id, oui_data=oui_data,
                            sw_resp=True)
                # check validity.
                if not utils.is_valid_vlan_tag(vdp_vlan):
                    LOG.error("Returned vlan %(vlan)s is invalid.",
                              {'vlan': vdp_vlan})
                    # Need to invoke CB. So no return here.
                    vdp_vlan = 0
                exist_vdp_vlan = lvdp_dict.get('vdp_vlan')
                exist_fail_reason = lvdp_dict.get('fail_reason')
                callback_count = lvdp_dict.get('callback_count')
                # Condition will be hit only during error cases when switch
                # reloads or when compute reloads
                if vdp_vlan != exist_vdp_vlan or (
                   fail_reason != exist_fail_reason or
                   callback_count > vdp_const.CALLBACK_THRESHOLD):
                    # Invoke the CB Function
                    cb_fn = lvdp_dict.get('vsw_cb_fn')
                    cb_data = lvdp_dict.get('vsw_cb_data')
                    if cb_fn:
                        cb_fn(cb_data, vdp_vlan, fail_reason)
                    lvdp_dict['vdp_vlan'] = vdp_vlan
                    lvdp_dict['fail_reason'] = fail_reason
                    lvdp_dict['callback_count'] = 0
                else:
                    lvdp_dict['callback_count'] += 1
        except Exception as e:
            LOG.error("Exception in Refrsh %s", str(e))

    def run_lldptool(self, args):
        """Function for invoking the lldptool utility. """
        full_args = ['lldptool'] + args
        try:
            utils.execute(full_args, root_helper=self.root_helper)
        except Exception as e:
            LOG.error("Unable to execute %(cmd)s. "
                      "Exception: %(exception)s",
                      {'cmd': full_args, 'exception': e})

    def store_oui(self, port_uuid, oui_type, oui_data):
        """Function for storing the OUI.

        param uuid: UUID of the vNIC
        param oui_type: OUI ID
        param oui_data: OUI Opaque Data
        """
        self.oui_vif_map[port_uuid] = {'oui_id': oui_type,
                                       'oui_data': oui_data}

    def store_vdp_vsi(self, port_uuid, mgrid, typeid, typeid_ver,
                      vsiid_frmt, vsiid, filter_frmt, gid, mac, vlan,
                      new_network, reply, oui_id, oui_data, vsw_cb_fn,
                      vsw_cb_data, reason):
        """Stores the vNIC specific info for VDP Refresh.

        :param uuid: vNIC UUID
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param vsiid: VSI value
        :param filter_frmt: Filter Format
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param new_network: Is this the first vNIC of this network
        :param reply: Response from the switch
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param vsw_cb_fn: Callback function from the app.
        :param vsw_cb_data: Callback data for the app.
        :param reason: Failure Reason
        """
        if port_uuid in self.vdp_vif_map:
            LOG.debug("Not Storing VDP VSI MAC %(mac)s UUID %(uuid)s",
                      {'mac': mac, 'uuid': vsiid})
        if new_network:
            vdp_vlan = reply
        else:
            vdp_vlan = vlan
        vdp_dict = {'vdp_vlan': vdp_vlan,
                    'mgrid': mgrid,
                    'typeid': typeid,
                    'typeid_ver': typeid_ver,
                    'vsiid_frmt': vsiid_frmt,
                    'vsiid': vsiid,
                    'filter_frmt': filter_frmt,
                    'mac': mac,
                    'gid': gid,
                    'vsw_cb_fn': vsw_cb_fn,
                    'vsw_cb_data': vsw_cb_data,
                    'fail_reason': reason,
                    'callback_count': 0}
        self.vdp_vif_map[port_uuid] = vdp_dict
        LOG.debug("Storing VDP VSI MAC %(mac)s UUID %(uuid)s VDP VLAN "
                  "%(vlan)s", {'mac': mac, 'uuid': vsiid, 'vlan': vdp_vlan})
        if oui_id:
            self.store_oui(port_uuid, oui_id, oui_data)

    def clear_oui(self, port_uuid):
        """Clears the OUI specific info.

        :param uuid: vNIC UUID

        Currently only one OUI per VSI fixme(padkrish)
        """
        if port_uuid in self.oui_vif_map:
            del self.oui_vif_map[port_uuid]
        else:
            LOG.debug("OUI does not exist")

    def clear_vdp_vsi(self, port_uuid):
        """Stores the vNIC specific info for VDP Refresh.

        :param uuid: vNIC UUID
        """
        try:
            LOG.debug("Clearing VDP VSI MAC %(mac)s UUID %(uuid)s",
                      {'mac': self.vdp_vif_map[port_uuid].get('mac'),
                       'uuid': self.vdp_vif_map[port_uuid].get('vsiid')})
            del self.vdp_vif_map[port_uuid]
        except Exception:
            LOG.error("VSI does not exist")
        self.clear_oui(port_uuid)

    def gen_cisco_vdp_oui(self, oui_id, oui_data):
        """Cisco specific handler for constructing OUI arguments. """
        oui_list = []
        vm_name = oui_data.get('vm_name')
        if vm_name is not None:
            oui_str = "oui=%s," % oui_id
            oui_name_str = oui_str + "vm_name=" + vm_name
            oui_list.append(oui_name_str)
        ip_addr = oui_data.get('ip_addr')
        if ip_addr is not None:
            oui_str = "oui=%s," % oui_id
            ip_addr_str = oui_str + "ipv4_addr=" + ip_addr
            oui_list.append(ip_addr_str)
        vm_uuid = oui_data.get('vm_uuid')
        if vm_uuid is not None:
            oui_str = "oui=%s," % oui_id
            vm_uuid_str = oui_str + "vm_uuid=" + vm_uuid
            oui_list.append(vm_uuid_str)
        return oui_list

    def gen_oui_str(self, oui_list):
        """Generate the OUI string for vdptool. """
        oui_str = []
        for oui in oui_list:
            oui_str.append('-c')
            oui_str.append(oui)
        return oui_str

    def construct_vdp_dict(self, mode, mgrid, typeid, typeid_ver, vsiid_frmt,
                           vsiid, filter_frmt, gid, mac, vlan, oui_id,
                           oui_data):
        """Constructs the VDP Message.

        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP
        Section for more detailed information
        :param mode: Associate or De-associate
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param vsiid: VSI value
        :param filter_frmt: Filter Format
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :return vdp_keyword_str: Dictionary of VDP arguments and values
        """
        vdp_keyword_str = {}
        if mgrid is None:
            mgrid = self.vdp_opts.get('mgrid')
        mgrid_str = "mgrid2=%s" % mgrid
        if typeid is None:
            typeid = self.vdp_opts.get('typeid')
        typeid_str = "typeid=%s" % typeid
        if typeid_ver is None:
            typeid_ver = self.vdp_opts.get('typeidver')
        typeid_ver_str = "typeidver=%s" % typeid_ver
        if int(vsiid_frmt) == int(self.vdp_opts.get('vsiidfrmt')):
            vsiid_str = "uuid=%s" % vsiid
        else:
            # Only format supported for now
            LOG.error("Unsupported VSIID Format1")
            return vdp_keyword_str
        if vlan == constants.INVALID_VLAN:
            vlan = 0
        if int(filter_frmt) == vdp_const.VDP_FILTER_GIDMACVID:
            if not mac or gid == 0:
                LOG.error("Incorrect Filter Format Specified")
                return vdp_keyword_str
            else:
                f = "filter=%s-%s-%s"
                filter_str = f % (vlan, mac, gid)
        elif int(filter_frmt) == vdp_const.VDP_FILTER_GIDVID:
            if gid == 0:
                LOG.error("NULL GID Specified")
                return vdp_keyword_str
            else:
                filter_str = "filter=" + '%d' % vlan + "--" + '%ld' % gid
        elif int(filter_frmt) == vdp_const.VDP_FILTER_MACVID:
            if not mac:
                LOG.error("NULL MAC Specified")
                return vdp_keyword_str
            else:
                filter_str = "filter=" + '%d' % vlan + "-" + mac
        elif int(filter_frmt) == vdp_const.VDP_FILTER_VID:
            filter_str = "filter=" + '%d' % vlan
        else:
            LOG.error("Incorrect Filter Format Specified")
            return vdp_keyword_str
        oui_list = []
        if oui_id is not None and oui_data is not None:
            if oui_id is 'cisco':
                oui_list = self.gen_cisco_vdp_oui(oui_id, oui_data)
        mode_str = "mode=" + mode
        vdp_keyword_str = dict(mode=mode_str, mgrid=mgrid_str,
                               typeid=typeid_str, typeid_ver=typeid_ver_str,
                               vsiid=vsiid_str, filter=filter_str,
                               oui_list=oui_list)
        return vdp_keyword_str

    def send_vdp_query_msg(self, mode, mgrid, typeid, typeid_ver, vsiid_frmt,
                           vsiid, filter_frmt, gid, mac, vlan, oui_id,
                           oui_data):
        """Constructs and Sends the VDP Query Message.

        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP
        Section for more detailed information
        :param mode: Associate or De-associate
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param vsiid: VSI value
        :param filter_frmt: Filter Format
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param sw_resp: Flag indicating if response is required from the daemon
        :return reply: Reply from vdptool
        """
        if not self.is_ncb:
            LOG.error("EVB cannot be set on NB")
            return
        vdp_key_str = self.construct_vdp_dict(mode, mgrid, typeid,
                                              typeid_ver, vsiid_frmt, vsiid,
                                              filter_frmt, gid, mac, vlan,
                                              None, None)
        if len(vdp_key_str) == 0:
            LOG.error("NULL List")
            return
        reply = self.run_vdptool(["-t", "-i", self.port_name, "-R", "-V", mode,
                                  "-c", vdp_key_str['mode'],
                                  "-c", vdp_key_str['mgrid'],
                                  "-c", vdp_key_str['typeid'],
                                  "-c", vdp_key_str['typeid_ver'],
                                  "-c", vdp_key_str['vsiid']])
        return reply

    def send_vdp_msg(self, mode, mgrid, typeid, typeid_ver, vsiid_frmt,
                     vsiid, filter_frmt, gid, mac, vlan, oui_id, oui_data,
                     sw_resp):
        """Constructs and Sends the VDP Message.

        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP
        Section for more detailed information
        :param mode: Associate or De-associate
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param vsiid: VSI value
        :param filter_frmt: Filter Format
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param sw_resp: Flag indicating if response is required from the daemon
        :return reply: Reply from vdptool
        """
        if not self.is_ncb:
            LOG.error("EVB cannot be set on NB")
            return
        vdp_key_str = self.construct_vdp_dict(mode, mgrid, typeid,
                                              typeid_ver, vsiid_frmt, vsiid,
                                              filter_frmt, gid, mac, vlan,
                                              oui_id, oui_data)
        if len(vdp_key_str) == 0:
            LOG.error("NULL List")
            return
        oui_cmd_str = self.gen_oui_str(vdp_key_str['oui_list'])
        if sw_resp:
            # If filter is not VID and if VLAN is 0, Query for the TLV first,
            # if found VDP will return the VLAN. Add support for this once
            # vdptool has the support for querying exact VSI filters
            # fixme(padkrish)

            reply = self.run_vdptool(["-T", "-i", self.port_name, "-W",
                                      "-V", mode, "-c", vdp_key_str['mode'],
                                      "-c", vdp_key_str['mgrid'], "-c",
                                      vdp_key_str['typeid'],
                                      "-c", vdp_key_str['typeid_ver'], "-c",
                                      vdp_key_str['vsiid'], "-c",
                                      "hints=none", "-c",
                                      vdp_key_str['filter']],
                                     oui_args=oui_cmd_str)
        else:
            reply = self.run_vdptool(["-T", "-i", self.port_name,
                                      "-V", mode, "-c", vdp_key_str['mode'],
                                      "-c", vdp_key_str['mgrid'], "-c",
                                      vdp_key_str['typeid'],
                                      "-c", vdp_key_str['typeid_ver'], "-c",
                                      vdp_key_str['vsiid'], "-c",
                                      "hints=none", "-c",
                                      vdp_key_str['filter']],
                                     oui_args=oui_cmd_str)
        return reply

    def crosscheck_reply_vsiid_mac(self, reply, vsiid, mac):
        """Cross Check the reply against the input vsiid,mac for associate. """
        vsiid_reply = reply.partition("uuid = ")[2].split()[0]
        if vsiid != vsiid_reply:
            fail_reason = vdp_const.vsi_mismatch_failure_reason % (
                vsiid, vsiid_reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        mac_reply = reply.partition("filter = ")[2].split('-')[1]
        if mac != mac_reply:
            fail_reason = vdp_const.mac_mismatch_failure_reason % (
                mac, mac_reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        return True, None

    def crosscheck_query_vsiid_mac(self, reply, vsiid, mac):
        """Cross Check the reply against the input vsiid,mac for get query. """
        vsiid_reply = reply.partition("uuid")[2].split()[0][4:]
        if vsiid != vsiid_reply:
            fail_reason = vdp_const.vsi_mismatch_failure_reason % (
                vsiid, vsiid_reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        mac_reply = reply.partition("filter")[2].split('-')[1]
        if mac != mac_reply:
            fail_reason = vdp_const.mac_mismatch_failure_reason % (
                mac, mac_reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        return True, None

    def get_vdp_failure_reason(self, reply):
        """Parse the failure reason from VDP. """
        try:
            fail_reason = reply.partition(
                "filter")[0].replace('\t', '').split('\n')[-2]
            if len(fail_reason) == 0:
                fail_reason = vdp_const.retrieve_failure_reason % (reply)
        except Exception:
            fail_reason = vdp_const.retrieve_failure_reason % (reply)
        return fail_reason

    def check_filter_validity(self, reply, filter_str):
        """Check for the validify of the filter. """
        try:
            f_ind = reply.index(filter_str)
            l_ind = reply.rindex(filter_str)
        except Exception:
            fail_reason = vdp_const.filter_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        if f_ind != l_ind:
            # Currently not supported if reply contains a filter keyword
            fail_reason = vdp_const.multiple_filter_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        return True, None

    def get_vlan_from_associate_reply(self, reply, vsiid, mac):
        """Parse the associate reply from VDP daemon to get the VLAN value. """
        try:
            verify_flag, fail_reason = self.crosscheck_reply_vsiid_mac(
                reply, vsiid, mac)
            if not verify_flag:
                return constants.INVALID_VLAN, fail_reason
            mode_str = reply.partition("mode = ")[2].split()[0]
            if mode_str != "assoc":
                fail_reason = self.get_vdp_failure_reason(reply)
                return constants.INVALID_VLAN, fail_reason
        except Exception:
            fail_reason = vdp_const.mode_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return constants.INVALID_VLAN, fail_reason
        check_filter, fail_reason = self.check_filter_validity(
            reply, "filter = ")
        if not check_filter:
            return constants.INVALID_VLAN, fail_reason
        try:
            vlan_val = reply.partition("filter = ")[2].split('-')[0]
            vlan = int(vlan_val)
        except ValueError:
            fail_reason = vdp_const.format_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return constants.INVALID_VLAN, fail_reason
        return vlan, None

    def check_hints(self, reply):
        """Parse the hints to check for errors. """
        try:
            f_ind = reply.index("hints")
            l_ind = reply.rindex("hints")
        except Exception:
            fail_reason = vdp_const.hints_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        if f_ind != l_ind:
            # Currently not supported if reply contains a filter keyword
            fail_reason = vdp_const.multiple_hints_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        try:
            hints_compl = reply.partition("hints")[2]
            hints_val = reply.partition("hints")[2][0:4]
            len_hints = int(hints_val)
            hints_val = hints_compl[4:4 + len_hints]
            hints = int(hints_val)
            if hints != 0:
                fail_reason = vdp_const.nonzero_hints_failure % (hints)
                return False, fail_reason
        except ValueError:
            fail_reason = vdp_const.format_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return False, fail_reason
        return True, None

    def get_vlan_from_query_reply(self, reply, vsiid, mac):
        """Parse the query reply from VDP daemon to get the VLAN value. """
        hints_ret, fail_reason = self.check_hints(reply)
        if not hints_ret:
            LOG.error("Incorrect hints found %s", reply)
            return constants.INVALID_VLAN, fail_reason
        check_filter, fail_reason = self.check_filter_validity(reply, "filter")
        if not check_filter:
            return constants.INVALID_VLAN, fail_reason
        try:
            verify_flag, fail_reason = self.crosscheck_query_vsiid_mac(
                reply, vsiid, mac)
            if not verify_flag:
                return constants.INVALID_VLAN, fail_reason
            filter_val = reply.partition("filter")[2]
            len_fil = len(filter_val)
            vlan_val = filter_val[4:len_fil].split('-')[0]
            vlan = int(vlan_val)
        except ValueError:
            fail_reason = vdp_const.format_failure_reason % (reply)
            LOG.error("%s", fail_reason)
            return constants.INVALID_VLAN, fail_reason
        return vlan, None

    def send_vdp_assoc(self, vsiid=None, mgrid=None, typeid=None,
                       typeid_ver=None, vsiid_frmt=vdp_const.VDP_VSIFRMT_UUID,
                       filter_frmt=vdp_const.VDP_FILTER_GIDMACVID, gid=0,
                       mac="", vlan=0, oui_id="", oui_data="", sw_resp=False):
        """Sends the VDP Associate Message.

        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP
        Section for more detailed information
        :param vsiid: VSI value, Only UUID supported for now
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param filter_frmt: Filter Format. Only <GID,MAC,VID> supported for now
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param sw_resp: Flag indicating if response is required from the daemon
        :return vlan: VLAN value returned by vdptool which in turn is given
        :             by Switch
        """

        if sw_resp and filter_frmt == vdp_const.VDP_FILTER_GIDMACVID:
            reply = self.send_vdp_query_msg("assoc", mgrid, typeid, typeid_ver,
                                            vsiid_frmt, vsiid, filter_frmt,
                                            gid, mac, vlan, oui_id, oui_data)
            vlan_resp, fail_reason = self.get_vlan_from_query_reply(
                reply, vsiid, mac)
            if vlan_resp != constants.INVALID_VLAN:
                return vlan_resp, fail_reason
        reply = self.send_vdp_msg("assoc", mgrid, typeid, typeid_ver,
                                  vsiid_frmt, vsiid, filter_frmt, gid, mac,
                                  vlan, oui_id, oui_data, sw_resp)
        if sw_resp:
            vlan, fail_reason = self.get_vlan_from_associate_reply(
                reply, vsiid, mac)
            return vlan, fail_reason
        return None, None

    def send_vdp_deassoc(self, vsiid=None, mgrid=None, typeid=None,
                         typeid_ver=None,
                         vsiid_frmt=vdp_const.VDP_VSIFRMT_UUID,
                         filter_frmt=vdp_const.VDP_FILTER_GIDMACVID, gid=0,
                         mac="", vlan=0, oui_id="", oui_data="",
                         sw_resp=False):
        """Sends the VDP Dis-Associate Message.

        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP Section
        for more detailed information
        :param vsiid: VSI value, Only UUID supported for now
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param filter_frmt: Filter Format. Only <GID,MAC,VID> supported for now
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param sw_resp: Flag indicating if response is required from the daemon
        """
        if filter_frmt == vdp_const.VDP_FILTER_GIDMACVID:
            reply = self.send_vdp_query_msg("assoc", mgrid, typeid, typeid_ver,
                                            vsiid_frmt, vsiid, filter_frmt,
                                            gid, mac, vlan, oui_id, oui_data)
            vlan_resp, fail_reason = self.get_vlan_from_query_reply(
                reply, vsiid, mac)
            # This is to cover cases where the enabler has a different VLAN
            # than LLDPAD. deassoc won't go through if wrong VLAN is passed.
            # Since enabler does not have right VLAN, most likely flows are not
            # programmed. Otherwise, there will be stale flows. No way of
            # knowing unless all flows are read and compared.
            if vlan_resp != constants.INVALID_VLAN:
                if vlan != vlan_resp:
                    LOG.info("vlan_resp %(resp)s different from passed "
                             "VLAN %(vlan)s", {'resp': vlan_resp,
                                               'vlan': vlan})
                    vlan = vlan_resp
        self.send_vdp_msg("deassoc", mgrid, typeid, typeid_ver,
                          vsiid_frmt, vsiid, filter_frmt, gid, mac, vlan,
                          oui_id, oui_data, sw_resp)

    def send_vdp_vnic_up(self, port_uuid=None, vsiid=None,
                         mgrid=None, typeid=None, typeid_ver=None,
                         vsiid_frmt=vdp_const.VDP_VSIFRMT_UUID,
                         filter_frmt=vdp_const.VDP_FILTER_GIDMACVID,
                         gid=0, mac="", vlan=0, oui=None,
                         new_network=False, vsw_cb_fn=None, vsw_cb_data=None):
        """Interface function to apps, called for a vNIC UP.

        This currently sends an VDP associate message.
        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP
        Section for more detailed information
        :param uuid: uuid of the vNIC
        :param vsiid: VSI value, Only UUID supported for now
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param filter_frmt: Filter Format. Only <GID,MAC,VID> supported for now
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param sw_resp: Flag indicating if response is required from the daemon
        :return reply: VLAN reply from vdptool
        """
        if oui is None:
            oui = {}
        oui_id = None
        oui_data = None
        if 'oui_id' in oui:
            oui_id = oui['oui_id']
            oui_data = oui
        reply, fail_reason = self.send_vdp_assoc(
            vsiid=vsiid, mgrid=mgrid, typeid=typeid, typeid_ver=typeid_ver,
            vsiid_frmt=vsiid_frmt, filter_frmt=filter_frmt, gid=gid, mac=mac,
            vlan=vlan, oui_id=oui_id, oui_data=oui_data, sw_resp=new_network)
        self.store_vdp_vsi(port_uuid, mgrid, typeid, typeid_ver,
                           vsiid_frmt, vsiid, filter_frmt, gid, mac, vlan,
                           new_network, reply, oui_id, oui_data, vsw_cb_fn,
                           vsw_cb_data, fail_reason)
        return reply, fail_reason

    def send_vdp_vnic_down(self, port_uuid=None, vsiid=None, mgrid=None,
                           typeid=None, typeid_ver=None,
                           vsiid_frmt=vdp_const.VDP_VSIFRMT_UUID,
                           filter_frmt=vdp_const.VDP_FILTER_GIDMACVID,
                           gid=0, mac="", vlan=0, oui=""):
        """Interface function to apps, called for a vNIC DOWN.

        This currently sends an VDP dis-associate message.
        Please refer http://www.ieee802.org/1/pages/802.1bg.html VDP
        Section for more detailed information
        :param uuid: uuid of the vNIC
        :param vsiid: VSI value, Only UUID supported for now
        :param mgrid: MGR ID
        :param typeid: Type ID
        :param typeid_ver: Version of the Type ID
        :param vsiid_frmt: Format of the following VSI argument
        :param filter_frmt: Filter Format. Only <GID,MAC,VID> supported for now
        :param gid: Group ID the vNIC belongs to
        :param mac: MAC Address of the vNIC
        :param vlan: VLAN of the vNIC
        :param oui_id: OUI Type
        :param oui_data: OUI Data
        :param sw_resp: Flag indicating if response is required from the daemon
        """
        # Correct non-zero VLAN needs to be specified
        try:
            with self.mutex_lock:
                self.send_vdp_deassoc(vsiid=vsiid, mgrid=mgrid, typeid=typeid,
                                      typeid_ver=typeid_ver,
                                      vsiid_frmt=vsiid_frmt,
                                      filter_frmt=filter_frmt, gid=gid,
                                      mac=mac, vlan=vlan)
                self.clear_vdp_vsi(port_uuid)
        except Exception as e:
            LOG.error("VNIC Down exception %s", e)

    def run_vdptool(self, args, oui_args=None):
        """Function that runs the vdptool utility. """
        if oui_args is None:
            oui_args = []
        full_args = ['vdptool'] + args + oui_args
        try:
            return utils.execute(full_args, root_helper=self.root_helper)
        except Exception as e:
            LOG.error("Unable to execute %(cmd)s. "
                      "Exception: %(exception)s",
                      {'cmd': full_args, 'exception': e})



# Copyright 2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ssl
import urllib2
from urllib2 import *  # noqa

from oslo_config import cfg


def build_opener(*handlers):
    if not cfg.CONF.ml2_cisco_ucsm.ucsm_https_verify:
        # SSL certificate checking has been turned off. Try to create
        # a default context.
        try:
            ctx = ssl.create_default_context()
        except Exception:
            # Python version does not support creation of default
            # context. In that case, proceed with the regular
            # build_opener.
            return urllib2.build_opener(*handlers)
        else:
            # Update successfully created context to ignore
            # SSL certificates on this HTTPS connection.
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
            return urllib2.build_opener(urllib2.HTTPSHandler(context=ctx),
                *handlers)
    else:
        # SSL certificate checking has not be turned off. Continue
        # using the unmodified build_opener().
        return urllib2.build_opener(*handlers)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils

from networking_cisco.plugins.cisco.cfg_agent import cfg_exceptions

LOG = logging.getLogger(__name__)


class DeviceDriverManager(object):
    """This class acts as a manager for device drivers.

    The device driver manager  maintains the relationship between the
    different neutron logical resource (eg: routers, firewalls, vpns etc.) and
    where they are hosted. For configuring a logical resource (router) in a
    hosting device, a corresponding device driver object is used.
    Device drivers encapsulate the necessary configuration information to
    configure a logical resource (eg: routers, firewalls, vpns etc.) on a
    hosting device (eg: ASR1k).

    The device driver class loads one driver object per hosting device.
    The loaded drivers are cached in memory, so when a request is made to
    get driver object for the same hosting device and resource (like router),
    the existing driver object is reused.

    This class is used by the service helper classes.
    """

    def __init__(self, cfg_agent):
        self._drivers = {}
        self._cfg_agent = cfg_agent
        self._hosting_device_routing_drivers_binding = {}

    def get_driver(self, resource_id):
        try:
            return self._drivers[resource_id]
        except KeyError:
            with excutils.save_and_reraise_exception(reraise=False):
                raise cfg_exceptions.DriverNotFound(resource='router',
                                                    id=resource_id)

    def get_driver_for_hosting_device(self, hd_id):
        try:
            return self._hosting_device_routing_drivers_binding[hd_id]
        except KeyError:
            with excutils.save_and_reraise_exception(reraise=False):
                raise cfg_exceptions.DriverNotFound(resource='hosting device',
                                                    id=hd_id)

    def set_driver(self, resource):
        """Set the driver for a neutron resource.

        :param resource: Neutron resource in dict format.
            Expected keys::

                {
                    'id': <value>,
                    'hosting_device': { 'id': <value>, },
                    'router_type': {'cfg_agent_driver': <value>,  }
                }

        :returns: driver object
        """
        try:
            resource_id = resource['id']
            hosting_device = resource['hosting_device']
            hd_id = hosting_device['id']
            if hd_id in self._hosting_device_routing_drivers_binding:
                driver = self._hosting_device_routing_drivers_binding[hd_id]
                self._drivers[resource_id] = driver
            else:
                driver_class = resource['router_type']['cfg_agent_driver']
                # save a copy of the obfuscated credentials
                obfusc_creds = dict(hosting_device.get('credentials'))
                if obfusc_creds:
                    # get un-obfuscated password
                    real_pw = self._cfg_agent.get_hosting_device_password(
                        obfusc_creds.get('credentials_id'))
                    hosting_device['credentials']['password'] = real_pw
                driver = importutils.import_object(driver_class,
                                                   **hosting_device)
                self._hosting_device_routing_drivers_binding[hd_id] = driver
                if obfusc_creds:
                    hosting_device['credentials'] = obfusc_creds
                self._drivers[resource_id] = driver
            return driver
        except ImportError:
            with excutils.save_and_reraise_exception(reraise=False):
                LOG.exception("Error loading cfg agent driver %(driver)s "
                              "for hosting device template %(t_name)s"
                              "(%(t_id)s)",
                              {'driver': driver_class, 't_id': hd_id,
                               't_name': resource['name']})
                raise cfg_exceptions.DriverNotExist(driver=driver_class)
        except KeyError as e:
            with excutils.save_and_reraise_exception(reraise=False):
                raise cfg_exceptions.DriverNotSetForMissingParameter(e)

    def remove_driver(self, resource_id):
        """Remove driver associated to a particular resource."""
        if resource_id in self._drivers:
            del self._drivers[resource_id]

    def remove_driver_for_hosting_device(self, hd_id):
        """Remove driver associated to a particular hosting device."""
        if hd_id in self._hosting_device_routing_drivers_binding:
            del self._hosting_device_routing_drivers_binding[hd_id]



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import mock
import netaddr
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.tests import base

from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_routing_driver as driver)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_snippets as snippets)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.iosxe import (
    cisco_iosxe_snippets as iosxe_snippets)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.iosxe import (
    iosxe_routing_driver as iosxe_driver)
from networking_cisco.plugins.cisco.cfg_agent.service_helpers import (
    routing_svc_helper)
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole
from networking_cisco.tests.unit.cisco.cfg_agent import cfg_agent_test_support

sys.modules['ncclient'] = mock.MagicMock()

_uuid = uuidutils.generate_uuid
DEV_NAME_LEN = iosxe_driver.IosXeRoutingDriver.DEV_NAME_LEN
HA_INFO = 'ha_info'
ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR
ROUTER_ROLE_HA_REDUNDANCY = cisco_constants.ROUTER_ROLE_HA_REDUNDANCY


class ASR1kRoutingDriver(base.BaseTestCase,
                         cfg_agent_test_support.CfgAgentTestSupportMixin):

    def setUp(self):
        super(ASR1kRoutingDriver, self).setUp()

        cfg.CONF.set_override('enable_multi_region', False, 'multi_region')

        device_params = self.prepare_hosting_device_params()
        self.driver = driver.ASR1kRoutingDriver(**device_params)
        self.driver._ncc_connection = mock.MagicMock()
        self.driver._check_response = mock.MagicMock(return_value=True)
        self.driver._check_acl = mock.MagicMock(return_value=False)

    def tearDown(self):
        super(ASR1kRoutingDriver, self).tearDown()
        self.driver._ncc_connection.reset_mock()

    def _create_test_routers(self, is_user_visible=True):
        self.router, ports = self.prepare_router_data(
            is_user_visible=is_user_visible)
        self.ri = routing_svc_helper.RouterInfo(self.router['id'],
                                                self.router)
        self.ha_priority = self.router[ha.DETAILS][ha.PRIORITY]
        self.vrf = ('nrouter-' + self.router['id'])[:DEV_NAME_LEN]

        # router port on external network, i.e., gateway port
        self.ext_gw_port = self.router['gw_port']
        self.ext_gw_port['ip_info'] = {
            'subnet_id': self.ext_gw_port['subnets'][0]['id'],
            'is_primary': True,
            'ip_cidr': self.ext_gw_port['subnets'][0]['cidr']
        }
        self.ext_phy_infc = (
            self.ext_gw_port['hosting_info']['physical_interface'])
        self.vlan_ext = self.ext_gw_port['hosting_info']['segmentation_id']
        self.ext_gw_upstream_ip = self.ext_gw_port['subnets'][0]['gateway_ip']
        self.ext_gw_ip = self.ext_gw_port['fixed_ips'][0]['ip_address']
        self.ext_gw_ip_cidr = self.ext_gw_port['subnets'][0]['cidr']
        self.ext_gw_ip_mask = str(
            netaddr.IPNetwork(self.ext_gw_ip_cidr).netmask)
        port_ha_info = self.ext_gw_port['ha_info']
        self.ext_gw_ha_group = port_ha_info['group']

        # router port on internal network
        self.int_port = ports[0]
        self.int_port['ip_info'] = {
            'subnet_id': self.int_port['subnets'][0]['id'],
            'is_primary': True,
            'ip_cidr': self.int_port['subnets'][0]['cidr']
        }
        self.int_port['change_details'] = {
            'new_ports': [self.int_port],
            'current_ports': [self.int_port],
            'old_ports': [],
            'former_ports': []
        }
        self.int_phy_infc = self.int_port['hosting_info']['physical_interface']
        self.vlan_int = self.int_port['hosting_info']['segmentation_id']
        self.int_gw_ip = self.int_port['fixed_ips'][0]['ip_address']
        self.int_gw_ip_cidr = self.int_port['subnets'][0]['cidr']
        self.int_gw_ip_mask = str(
            netaddr.IPNetwork(self.int_gw_ip_cidr).netmask)
        port_ha_info = self.int_port['ha_info']
        self.int_gw_ip_vip = (
            port_ha_info['ha_port']['fixed_ips'][0]['ip_address'])
        self.int_gw_ha_group = port_ha_info['group']
        self.floating_ip = '19.4.0.6'
        self.fixed_ip = '35.4.0.20'

    def _create_test_global_routers(self, num_ext_subnets=1, subnet_index=0):
        # global router and its ports
        self.global_router, gl_ports = self.prepare_router_data(
            is_global=True, num_ext_subnets=num_ext_subnets)
        self.ha_priority = self.global_router[ha.DETAILS][ha.PRIORITY]
        self.ri_global = routing_svc_helper.RouterInfo(
            self.global_router['id'], self.global_router)
        self.gl_port = gl_ports[0]
        self.gl_port['ip_info'] = {
            'subnet_id': self.gl_port['subnets'][0]['id'],
            'is_primary': True,
            'ip_cidr': self.gl_port['subnets'][0]['cidr']
        }
        self.ext_phy_infc = self.gl_port['hosting_info']['physical_interface']
        self.vlan_ext = self.gl_port['hosting_info']['segmentation_id']
        self.gl_port_ip = self.gl_port['fixed_ips'][subnet_index]['ip_address']
        self.gl_port_ip_cidr = self.gl_port['subnets'][subnet_index]['cidr']
        self.gl_port_ip_mask = str(
            netaddr.IPNetwork(self.gl_port_ip_cidr).netmask)
        port_ha_info = self.gl_port['ha_info']
        self.gl_port_vip = (
            port_ha_info['ha_port']['fixed_ips'][subnet_index]['ip_address'])
        self.gl_port_ha_group = port_ha_info['group']

    def assert_edit_run_cfg(self, snippet_name, args):
        if args:
            confstr = snippet_name % args
        else:
            confstr = snippet_name
        self.driver._ncc_connection.edit_config.assert_any_call(
            target='running', config=confstr)

    def _assert_number_of_edit_run_cfg_calls(self, num):
        self.assertEqual(num,
                         self.driver._ncc_connection.edit_config.call_count)

    def _generate_hsrp_cfg_args(self, subintfc, group, priority, vip, vlan):
        return (subintfc,
                group, priority,
                group, vip,
                group,
                group, group, vlan)

    def test_internal_network_added(self):
        self._create_test_routers()
        self.driver.internal_network_added(self.ri, self.int_port)

        sub_interface = self.int_phy_infc + '.' + str(self.vlan_int)
        cfg_args_sub = (sub_interface, self.vlan_int, self.vrf, self.int_gw_ip,
                        self.int_gw_ip_mask)
        self.assert_edit_run_cfg(
            snippets.CREATE_SUBINTERFACE_WITH_ID, cfg_args_sub)

        cfg_args_hsrp = self._generate_hsrp_cfg_args(
            sub_interface, self.int_gw_ha_group, self.ha_priority,
            self.int_gw_ip_vip, self.vlan_int)
        self.assert_edit_run_cfg(
            snippets.SET_INTC_ASR_HSRP_EXTERNAL, cfg_args_hsrp)

    def test_internal_network_added_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)

        region_id = cfg.CONF.multi_region.region_id

        vrf = self.vrf + "-" + region_id

        self.driver.internal_network_added(self.ri, self.int_port)

        sub_interface = self.int_phy_infc + '.' + str(self.vlan_int)
        cfg_args_sub = (sub_interface, region_id, self.vlan_int, vrf,
                        self.int_gw_ip, self.int_gw_ip_mask)
        self.assert_edit_run_cfg(
            snippets.CREATE_SUBINTERFACE_REGION_ID_WITH_ID, cfg_args_sub)

        cfg_args_hsrp = self._generate_hsrp_cfg_args(
            sub_interface, self.int_gw_ha_group, self.ha_priority,
            self.int_gw_ip_vip, self.vlan_int)
        self.assert_edit_run_cfg(
            snippets.SET_INTC_ASR_HSRP_EXTERNAL, cfg_args_hsrp)

    def test_internal_network_added_global_router(self):
        self._create_test_global_routers()
        self.driver.internal_network_added(self.ri_global, self.gl_port)
        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_args_sub = (sub_interface, self.vlan_ext,
                        self.gl_port_ip, self.gl_port_ip_mask)
        self.assert_edit_run_cfg(
            snippets.CREATE_SUBINTERFACE_EXTERNAL_WITH_ID, cfg_args_sub)

        cfg_args_hsrp = self._generate_hsrp_cfg_args(
            sub_interface, self.gl_port_ha_group, self.ha_priority,
            self.gl_port_vip, self.vlan_ext)
        self.assert_edit_run_cfg(
            snippets.SET_INTC_ASR_HSRP_EXTERNAL, cfg_args_hsrp)

    def test_internal_network_added_global_router_secondary_subnet(self):
        self._create_test_global_routers(num_ext_subnets=2, subnet_index=1)
        self.gl_port['ip_info']['subnet_id'] = self.gl_port['subnets'][1]['id']
        self.gl_port['ip_info']['ip_cidr'] = self.gl_port['subnets'][1]['cidr']
        self.gl_port['ip_info']['is_primary'] = False

        self.driver.internal_network_added(self.ri_global, self.gl_port)
        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_args_sub = (sub_interface, self.gl_port_ip, self.gl_port_ip_mask)
        self.assert_edit_run_cfg(
            snippets.SET_INTERFACE_SECONDARY_IP, cfg_args_sub)

        cfg_args_hsrp = (sub_interface, self.gl_port_ha_group,
                         self.gl_port_vip)
        self.assert_edit_run_cfg(
            snippets.SET_INTC_ASR_SECONDARY_HSRP_EXTERNAL, cfg_args_hsrp)

    def test_internal_network_added_global_router_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_global_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id

        self.driver.internal_network_added(self.ri_global, self.gl_port)
        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_args_sub = (sub_interface, region_id, self.vlan_ext,
                        self.gl_port_ip, self.gl_port_ip_mask)
        self.assert_edit_run_cfg(
            snippets.CREATE_SUBINTERFACE_EXT_REGION_ID_WITH_ID, cfg_args_sub)

        cfg_args_hsrp = self._generate_hsrp_cfg_args(
            sub_interface, self.gl_port_ha_group, self.ha_priority,
            self.gl_port_vip, self.vlan_ext)
        self.assert_edit_run_cfg(
            snippets.SET_INTC_ASR_HSRP_EXTERNAL, cfg_args_hsrp)

    def test_internal_network_added_global_router_with_multi_region_sec_sn(
            self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_global_routers(num_ext_subnets=2, subnet_index=1)
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)

        self.gl_port['ip_info']['subnet_id'] = self.gl_port['subnets'][1]['id']
        self.gl_port['ip_info']['ip_cidr'] = self.gl_port['subnets'][1]['cidr']
        self.gl_port['ip_info']['is_primary'] = False
        self.driver.internal_network_added(self.ri_global, self.gl_port)
        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_args_sub = (sub_interface, self.gl_port_ip, self.gl_port_ip_mask)
        self.assert_edit_run_cfg(
            snippets.SET_INTERFACE_SECONDARY_IP, cfg_args_sub)

        cfg_args_hsrp = (sub_interface, self.gl_port_ha_group,
                         self.gl_port_vip)
        self.assert_edit_run_cfg(
            snippets.SET_INTC_ASR_SECONDARY_HSRP_EXTERNAL, cfg_args_hsrp)

    def _make_test_router_non_ha(self):
        self._create_test_routers()
        self.ri.router[ha.ENABLED] = False
        del self.ri.router[ha.DETAILS]
        del self.ext_gw_port[HA_INFO]
        del self.int_port[HA_INFO]

    def test_external_network_added_non_ha(self):
        self._make_test_router_non_ha()
        self.driver.external_gateway_added(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.ENABLE_INTF, sub_interface)

        cfg_params_nat = (self.vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.CREATE_NAT_POOL, cfg_params_nat)

    def test_external_network_added_user_visible_router(self):
        self._create_test_routers()
        self.driver.external_gateway_added(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.ENABLE_INTF, sub_interface)

        cfg_params_nat = (self.vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.CREATE_NAT_POOL, cfg_params_nat)

    def test_external_network_added_redundancy_router(self):
        self._create_test_routers(is_user_visible=False)
        self.driver.external_gateway_added(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.ENABLE_INTF, sub_interface)

        cfg_params_nat = (self.vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.CREATE_NAT_POOL, cfg_params_nat)

    def test_external_network_added_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id
        vrf = self.vrf + "-" + region_id

        self.driver.external_gateway_added(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.ENABLE_INTF, sub_interface)

        cfg_params_nat = (vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.CREATE_NAT_POOL, cfg_params_nat)

    def test_external_gateway_removed_non_ha(self):
        self._make_test_router_non_ha()
        self.driver.external_gateway_removed(self.ri, self.ext_gw_port)

        cfg_params_nat = (self.vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.DELETE_NAT_POOL, cfg_params_nat)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_params_remove_route = (self.vrf,
                                   sub_interface, self.ext_gw_upstream_ip)
        self.assert_edit_run_cfg(snippets.REMOVE_DEFAULT_ROUTE_WITH_INTF,
                                 cfg_params_remove_route)

    def test_external_gateway_removed_user_visible_router(self):
        self._create_test_routers()
        self.driver.external_gateway_removed(self.ri, self.ext_gw_port)

        cfg_params_nat = (self.vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.DELETE_NAT_POOL, cfg_params_nat)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_params_remove_route = (self.vrf,
                                   sub_interface, self.ext_gw_upstream_ip)
        self.assert_edit_run_cfg(snippets.REMOVE_DEFAULT_ROUTE_WITH_INTF,
                                 cfg_params_remove_route)

    def test_external_gateway_removed_redundancy_router(self):
        self._create_test_routers(is_user_visible=False)
        self.driver.external_gateway_removed(self.ri, self.ext_gw_port)

        cfg_params_nat = (self.vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.DELETE_NAT_POOL, cfg_params_nat)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_params_remove_route = (self.vrf,
                                   sub_interface, self.ext_gw_upstream_ip)
        self.assert_edit_run_cfg(snippets.REMOVE_DEFAULT_ROUTE_WITH_INTF,
                                 cfg_params_remove_route)

    def test_external_gateway_removed_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id
        vrf = self.vrf + "-" + region_id

        self.driver.external_gateway_removed(self.ri, self.ext_gw_port)

        cfg_params_nat = (vrf + '_nat_pool', self.ext_gw_ip,
                          self.ext_gw_ip, self.ext_gw_ip_mask)
        self.assert_edit_run_cfg(snippets.DELETE_NAT_POOL, cfg_params_nat)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        cfg_params_remove_route = (vrf,
                                   sub_interface, self.ext_gw_upstream_ip)
        self.assert_edit_run_cfg(snippets.REMOVE_DEFAULT_ROUTE_WITH_INTF,
                                 cfg_params_remove_route)

    def test_external_gateway_removed_global_router(self):
        self._create_test_global_routers()
        self.driver._interface_exists = mock.MagicMock(return_value=True)

        self.driver.external_gateway_removed(self.ri_global, self.gl_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(
            iosxe_snippets.REMOVE_SUBINTERFACE, sub_interface)

    def test_floating_ip_added(self):
        self._create_test_routers()
        self.driver.floating_ip_added(self.ri, self.ext_gw_port,
                                      self.floating_ip, self.fixed_ip)

        self._assert_number_of_edit_run_cfg_calls(1)
        cfg_params_floating = (self.fixed_ip, self.floating_ip, self.vrf,
                               self.ext_gw_ha_group, self.vlan_ext)
        self.assert_edit_run_cfg(snippets.SET_STATIC_SRC_TRL_NO_VRF_MATCH,
                                 cfg_params_floating)

    def test_floating_ip_added_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id
        vrf = self.vrf + "-" + region_id

        self.driver.floating_ip_added(self.ri, self.ext_gw_port,
                                      self.floating_ip, self.fixed_ip)

        self._assert_number_of_edit_run_cfg_calls(1)
        cfg_params_floating = (self.fixed_ip, self.floating_ip, vrf,
                               self.ext_gw_ha_group, self.vlan_ext)
        self.assert_edit_run_cfg(snippets.SET_STATIC_SRC_TRL_NO_VRF_MATCH,
                                 cfg_params_floating)

    def test_floating_ip_removed(self):
        self._create_test_routers()
        self.driver.floating_ip_removed(self.ri, self.ext_gw_port,
                                        self.floating_ip, self.fixed_ip)

        self._assert_number_of_edit_run_cfg_calls(1)
        cfg_params_floating = (self.fixed_ip, self.floating_ip, self.vrf,
                               self.ext_gw_ha_group, self.vlan_ext)
        self.assert_edit_run_cfg(snippets.REMOVE_STATIC_SRC_TRL_NO_VRF_MATCH,
                                 cfg_params_floating)

    def test_floating_ip_removed_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id
        vrf = self.vrf + "-" + region_id

        self.driver.floating_ip_removed(self.ri, self.ext_gw_port,
                                      self.floating_ip, self.fixed_ip)

        self._assert_number_of_edit_run_cfg_calls(1)
        cfg_params_floating = (self.fixed_ip, self.floating_ip, vrf,
                               self.ext_gw_ha_group, self.vlan_ext)
        self.assert_edit_run_cfg(snippets.REMOVE_STATIC_SRC_TRL_NO_VRF_MATCH,
                                 cfg_params_floating)

    def test_driver_enable_internal_network_NAT(self):
        self._create_test_routers()
        self.driver.enable_internal_network_NAT(self.ri, self.int_port,
                                                self.ext_gw_port)

        self._assert_number_of_edit_run_cfg_calls(4)

        acl_name = '%(acl_prefix)s_%(vlan)s_%(port)s' % {
                   'acl_prefix': 'neutron_acl',
                   'vlan': self.vlan_int,
                   'port': self.int_port['id'][:8]}
        net = netaddr.IPNetwork(self.int_gw_ip_cidr).network
        net_mask = netaddr.IPNetwork(self.int_gw_ip_cidr).hostmask
        cfg_params_create_acl = (acl_name, net, net_mask)
        self.assert_edit_run_cfg(
            iosxe_snippets.CREATE_ACL, cfg_params_create_acl)

        pool_name = "%s_nat_pool" % self.vrf
        cfg_params_dyn_trans = (acl_name, pool_name, self.vrf)
        self.assert_edit_run_cfg(
            snippets.SET_DYN_SRC_TRL_POOL, cfg_params_dyn_trans)

        sub_interface_int = self.int_phy_infc + '.' + str(self.vlan_int)
        sub_interface_ext = self.int_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.SET_NAT,
                                 (sub_interface_int, 'inside'))
        self.assert_edit_run_cfg(iosxe_snippets.SET_NAT,
                                 (sub_interface_ext, 'outside'))

    def test_driver_enable_internal_network_NAT_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id
        vrf = self.vrf + "-" + region_id

        self.driver.enable_internal_network_NAT(self.ri, self.int_port,
                                                self.ext_gw_port)

        self._assert_number_of_edit_run_cfg_calls(4)

        acl_name = '%(acl_prefix)s_%(region_id)s_%(vlan)s_%(port)s' % {
                   'acl_prefix': 'neutron_acl',
                   'region_id': region_id,
                   'vlan': self.vlan_int,
                   'port': self.int_port['id'][:8]}

        net = netaddr.IPNetwork(self.int_gw_ip_cidr).network
        net_mask = netaddr.IPNetwork(self.int_gw_ip_cidr).hostmask
        cfg_params_create_acl = (acl_name, net, net_mask)
        self.assert_edit_run_cfg(
            iosxe_snippets.CREATE_ACL, cfg_params_create_acl)

        pool_name = "%s_nat_pool" % vrf
        cfg_params_dyn_trans = (acl_name, pool_name, vrf)
        self.assert_edit_run_cfg(
            snippets.SET_DYN_SRC_TRL_POOL, cfg_params_dyn_trans)

        sub_interface_int = self.int_phy_infc + '.' + str(self.vlan_int)
        sub_interface_ext = self.int_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.SET_NAT,
                                 (sub_interface_int, 'inside'))
        self.assert_edit_run_cfg(iosxe_snippets.SET_NAT,
                                 (sub_interface_ext, 'outside'))

    def test_driver_disable_internal_network_NAT(self):
        self._create_test_routers()
        self.driver.disable_internal_network_NAT(self.ri, self.int_port,
                                                 self.ext_gw_port)

        self._assert_number_of_edit_run_cfg_calls(3)

        acl_name = '%(acl_prefix)s_%(vlan)s_%(port)s' % {
                   'acl_prefix': 'neutron_acl',
                   'vlan': self.vlan_int,
                   'port': self.int_port['id'][:8]}
        pool_name = "%s_nat_pool" % self.vrf

        cfg_params_dyn_trans = (acl_name, pool_name, self.vrf)
        self.assert_edit_run_cfg(
            snippets.REMOVE_DYN_SRC_TRL_POOL, cfg_params_dyn_trans)

        self.assert_edit_run_cfg(iosxe_snippets.REMOVE_ACL, acl_name)

    def test_driver_disable_internal_network_NAT_with_multi_region(self):
        cfg.CONF.set_override('enable_multi_region', True, 'multi_region')
        self._create_test_routers()
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        self.assertEqual(True, is_multi_region_enabled)
        region_id = cfg.CONF.multi_region.region_id
        vrf = self.vrf + "-" + region_id

        self.driver.disable_internal_network_NAT(self.ri, self.int_port,
                                                 self.ext_gw_port)

        self._assert_number_of_edit_run_cfg_calls(3)

        acl_name = '%(acl_prefix)s_%(region_id)s_%(vlan)s_%(port)s' % {
                   'acl_prefix': 'neutron_acl',
                   'region_id': region_id,
                   'vlan': self.vlan_int,
                   'port': self.int_port['id'][:8]}

        pool_name = "%s_nat_pool" % vrf

        cfg_params_dyn_trans = (acl_name, pool_name, vrf)
        self.assert_edit_run_cfg(
            snippets.REMOVE_DYN_SRC_TRL_POOL, cfg_params_dyn_trans)

        self.assert_edit_run_cfg(iosxe_snippets.REMOVE_ACL, acl_name)

    def test_enable_interface_user_visible_router(self):
        self._create_test_routers()
        self.driver.enable_router_interface(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.ENABLE_INTF, sub_interface)

    def test_enable_interface_redundancy_router(self):
        self._create_test_routers(is_user_visible=False)
        self.driver.enable_router_interface(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.ENABLE_INTF, sub_interface)

    def test_disable_interface_user_visible_router(self):
        self._create_test_routers()
        self.driver.disable_router_interface(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.DISABLE_INTF, sub_interface)

    def test_disable_interface_redundancy_router(self):
        self._create_test_routers(is_user_visible=False)
        self.driver.disable_router_interface(self.ri, self.ext_gw_port)

        sub_interface = self.ext_phy_infc + '.' + str(self.vlan_ext)
        self.assert_edit_run_cfg(iosxe_snippets.DISABLE_INTF, sub_interface)

    def test_get_configuration(self):
        self._create_test_routers()
        self.driver._get_running_config = mock.MagicMock()
        self.driver.get_configuration()
        self.driver._get_running_config.assert_called_once_with(split=False)
